"""Put a node into standby mode and check that resources migrate."""

__all__ = ["StandbyTest"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.starttest import StartTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class StandbyTest(CTSTest):
    """Put a node into standby and check that resources migrate away from it."""

    def __init__(self, cm):
        """
        Create a new StandbyTest instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.benchmark = True
        self.name = "Standby"

        self._start = StartTest(cm)
        self._startall = SimulStartLite(cm)

    # make sure the node is active
    # set the node to standby mode
    # check resources, none resource should be running on the node
    # set the node to active mode
    # check resources, resources should have been migrated back (SHOULD THEY?)

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")
        ret = self._startall(None)
        if not ret:
            return self.failure("Start all nodes failed")

        self.debug(f"Make sure node {node} is active")
        if self._cm.in_standby_mode(node):
            if not self._cm.set_standby_mode(node, False):
                return self.failure(f"can't set node {node} to active mode")

        self._cm.cluster_stable()

        if self._cm.in_standby_mode(node):
            return self.failure(f"standby status of {node} is [on] but we expect [off]")

        watchpats = [
            r"State transition .* -> S_POLICY_ENGINE",
        ]
        watch = self.create_watch(watchpats, self._env["DeadTime"] + 10)
        watch.set_watch()

        self.debug(f"Setting node {node} to standby mode")
        if not self._cm.set_standby_mode(node, True):
            return self.failure(f"can't set node {node} to standby mode")

        self.set_timer("on")

        ret = watch.look_for_all()
        if not ret:
            self._logger.log(f"Patterns not found: {watch.unmatched!r}")
            self._cm.set_standby_mode(node, False)
            return self.failure(f"cluster didn't react to standby change on {node}")

        self._cm.cluster_stable()

        if not self._cm.in_standby_mode(node):
            return self.failure(f"standby status of {node} is [off] but we expect [on]")

        self.log_timer("on")

        self.debug("Checking resources")
        rscs_on_node = self._cm.active_resources(node)
        if rscs_on_node:
            rc = self.failure(f"{node} set to standby, {rscs_on_node!r} is still running on it")
            self.debug(f"Setting node {node} to active mode")
            self._cm.set_standby_mode(node, False)
            return rc

        self.debug(f"Setting node {node} to active mode")
        if not self._cm.set_standby_mode(node, False):
            return self.failure(f"can't set node {node} to active mode")

        self.set_timer("off")
        self._cm.cluster_stable()

        if self._cm.in_standby_mode(node):
            return self.failure(f"standby status of {node} is [on] but we expect [off]")

        self.log_timer("off")

        return self.success()



"""A module providing XML validation utilities."""

__copyright__ = "Copyright 2024-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+)"

__all__ = ["validate"]

import os
import subprocess
import sys

from pacemaker._cts.errors import XmlValidationError
from pacemaker._cts.process import pipe_communicate
from pacemaker.buildoptions import BuildOptions


def find_validator(rng_file):
    """
    Return the command line used to validate XML output.

    If no validator is found, return None.
    """
    if os.access(BuildOptions.XMLLINT_PATH, os.X_OK):
        if rng_file is None:
            return [BuildOptions.XMLLINT_PATH, "-"]

        return [BuildOptions.XMLLINT_PATH, "--relaxng", rng_file, "-"]

    raise FileNotFoundError(f"Could not find validator for {rng_file}")


def rng_directory():
    """Return the directory containing RNG schema files."""
    if os.environ.get("PCMK_schema_directory", "") != "":
        return os.environ["PCMK_schema_directory"]

    for path in sys.path:
        if os.path.exists(f"{path}/cts-fencing.in"):
            return f"{path}/../xml"

    return BuildOptions.SCHEMA_DIR


def validate(xml, check_rng=True, verbose=False):
    """Validate the given XML input string against a schema."""
    if check_rng:
        rng_file = f"{rng_directory()}/api/api-result.rng"
    else:
        rng_file = None

    cmd = find_validator(rng_file)

    if verbose:
        s = " ".join(cmd)
        print(f"\nRunning: {s}")

    with subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as validator:
        output = pipe_communicate(validator, check_stderr=True, stdin=xml)

        if validator.returncode != 0:
            raise XmlValidationError(output, validator.returncode)

        return output



"""Kill a pacemaker daemon and test how the cluster recovers."""

__all__ = ["ComponentFail"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import re

from pacemaker._cts.audits import AuditResource
from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._rsh is callable.
# pylint: disable=not-callable
# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


# @TODO Separate this into a separate test for each component, so the patterns
# can be made specific to each component, investigating failures is a little
# easier, and specific testing can be done for each component (for example,
# set attributes before and after killing pacemaker-attrd and check values).
class ComponentFail(CTSTest):
    """Kill a random pacemaker daemon and wait for the cluster to recover."""

    def __init__(self, cm):
        """
        Create a new ComponentFail instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.is_unsafe = True
        self.name = "ComponentFail"

        self._complist = cm.components
        self._okerrpatterns = []
        self._patterns = []
        self._startall = SimulStartLite(cm)

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")
        self._patterns = []
        self._okerrpatterns = []

        # start all nodes
        ret = self._startall(None)
        if not ret:
            return self.failure("Setup failed")

        if not self._cm.cluster_stable(self._env["StableTime"]):
            return self.failure("Setup failed - unstable")

        node_is_dc = self._cm.is_node_dc(node, None)

        # select a component to kill
        chosen = self._env.random_gen.choice(self._complist)
        while chosen.dc_only and not node_is_dc:
            chosen = self._env.random_gen.choice(self._complist)

        self.debug(f"...component {chosen.name} (dc={node_is_dc})")
        self.incr(chosen.name)

        if chosen.name != "corosync":
            self._patterns.extend([
                self.templates["Pat:ChildKilled"] % (node, chosen.name),
                self.templates["Pat:ChildRespawn"] % (node, chosen.name),
            ])

        self._patterns.extend(chosen.pats)
        if node_is_dc:
            self._patterns.extend(chosen.dc_pats)

        # @TODO this should be a flag in the Component
        if chosen.name in ["corosync", "pacemaker-based", "pacemaker-fenced"]:
            # Ignore actions for fence devices if fencer will respawn
            # (their registration will be lost, and probes will fail)
            self._okerrpatterns = [
                self.templates["Pat:Fencing_active"],
            ]
            (_, lines) = self._rsh(node, "crm_resource -c", verbose=1)

            for line in lines:
                if re.search("^Resource", line):
                    r = AuditResource(self._cm, line)

                    if r.rclass == "stonith":
                        self._okerrpatterns.extend([
                            self.templates["Pat:Fencing_recover"] % r.id,
                            self.templates["Pat:Fencing_probe"] % r.id,
                        ])

        # supply a copy so self.patterns doesn't end up empty
        tmp_pats = self._patterns.copy()
        self._patterns.extend(chosen.badnews_ignore)

        # Look for STONITH ops, depending on Env["at-boot"] we might need to change the nodes status
        stonith_pats = [
            self.templates["Pat:Fencing_ok"] % node
        ]
        stonith = self.create_watch(stonith_pats, 0)
        stonith.set_watch()

        # set the watch for stable
        watch = self.create_watch(
            tmp_pats, self._env["DeadTime"] + self._env["StableTime"] + self._env["StartTime"])

        watch.set_watch()

        # kill the component
        chosen.kill(node)

        self.debug("Waiting for the cluster to recover")
        self._cm.cluster_stable()

        self.debug("Waiting for any fenced node to come back up")
        self._cm.ns.wait_for_all_nodes(self._env["nodes"], 600)

        self.debug("Waiting for the cluster to re-stabilize with all nodes")
        self._cm.cluster_stable(self._env["StartTime"])

        self.debug(f"Checking if {node} was shot")
        shot = stonith.look(60)

        if shot:
            self.debug(f"Found: {shot!r}")
            self._okerrpatterns.append(self.templates["Pat:Fencing_start"] % node)

            if not self._env["at-boot"]:
                self._cm.expected_status[node] = "down"

            # If fencing occurred, chances are many (if not all) the expected logs
            # will not be sent - or will be lost when the node reboots
            return self.success()

        # check for logs indicating a graceful recovery
        matched = watch.look_for_all(allow_multiple_matches=True)
        if watch.unmatched:
            self._logger.log(f"Patterns not found: {watch.unmatched!r}")

        self.debug("Waiting for the cluster to re-stabilize with all nodes")
        is_stable = self._cm.cluster_stable(self._env["StartTime"])

        if not matched:
            return self.failure(f"Didn't find all expected {chosen.name} patterns")

        if not is_stable:
            return self.failure(f"Cluster did not become stable after killing {chosen.name}")

        return self.success()

    @property
    def errors_to_ignore(self):
        """Return a list of errors which should be ignored."""
        # Note that okerrpatterns refers to the last time we ran this test
        # The good news is that this works fine for us...
        self._okerrpatterns.extend(self._patterns)
        return self._okerrpatterns



"""Internal Python API for the `pacemaker` package."""

__copyright__ = "Copyright 2023-2024 the Pacemaker project contributors"
__license__ = "GNU Lesser General Public License version 2.1 or later (LGPLv2.1+)"



"""Timer-related utilities for CTS."""

__all__ = ["Timer"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import time


class Timer:
    """
    A class for measuring the runtime of some task.

    A Timer may be used manually or as a context manager, like so:

        with Timer(logger, "SomeTest", "SomeTimer"):
            ...

    A Timer runs from when start() is called until the timer is deleted or
    reset() is called.  There is no explicit stop method.
    """

    def __init__(self, logger, test_name, timer_name):
        """
        Create a new Timer instance.

        Arguments:
        logger      -- A Logger instance that can be used to record when
                       the timer stopped
        test_name   -- The name of the test this timer is being run for
        timer_name  -- The name of this timer
        """
        self._logger = logger
        self._start_time = None
        self._test_name = test_name
        self._timer_name = timer_name

    def __enter__(self):
        """When used as a context manager, start the timer."""
        self.start()
        return self

    def __exit__(self, *args):
        """When used as a context manager, log the elapsed time."""
        self._logger.debug(f"{self._test_name}:{self._timer_name} runtime: {self.elapsed:.2f}")

    def reset(self):
        """Restart the timer."""
        self.start()

    def start(self):
        """Start the timer."""
        self._start_time = time.time()

    @property
    def start_time(self):
        """Return when the timer started."""
        return self._start_time

    @property
    def elapsed(self):
        """Return how long the timer has been running for."""
        return time.time() - self._start_time



"""Start the cluster without a CIB and verify it gets copied from another node."""

__all__ = ["ResyncCIB"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker import BuildOptions
from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.restarttest import RestartTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.simulstoplite import SimulStopLite

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._rsh is callable.
# pylint: disable=not-callable


class ResyncCIB(CTSTest):
    """Start the cluster on a node without a CIB and verify the CIB is copied over later."""

    def __init__(self, cm):
        """
        Create a new ResyncCIB instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.name = "ResyncCIB"

        self._restart1 = RestartTest(cm)
        self._startall = SimulStartLite(cm)
        self._stopall = SimulStopLite(cm)

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")

        # Shut down all the nodes...
        if not self._stopall(None):
            return self.failure("Could not stop all nodes")

        # Test config recovery when the other nodes come up
        self._rsh(node, f"rm -f {BuildOptions.CIB_DIR}/cib*")

        # Start the selected node
        if not self._restart1(node):
            return self.failure(f"Could not start {node}")

        # Start all remaining nodes
        if not self._startall(None):
            return self.failure("Could not start the remaining nodes")

        return self.success()

    @property
    def errors_to_ignore(self):
        """Return a list of errors which should be ignored."""
        # Errors that occur as a result of the CIB being wiped
        return [
            r"error.*: v1 patchset error, patch failed to apply: Application of an update diff failed",
            r"error.*: Resource start-up disabled since no STONITH resources have been defined",
            r"error.*: Either configure some or disable STONITH with the stonith-enabled option",
            r"error.*: NOTE: Clusters with shared data need STONITH to ensure data integrity"
        ]



"""Start a node and then tell it to stop before it is fully running."""

__all__ = ["PartialStart"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.simulstoplite import SimulStopLite
from pacemaker._cts.tests.stoptest import StopTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class PartialStart(CTSTest):
    """Interrupt a node before it's finished starting up."""

    def __init__(self, cm):
        """
        Create a new PartialStart instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.name = "PartialStart"

        self._startall = SimulStartLite(cm)
        self._stop = StopTest(cm)
        self._stopall = SimulStopLite(cm)

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")

        ret = self._stopall(None)
        if not ret:
            return self.failure("Setup failed")

        watchpats = [
            "pacemaker-controld.*Connecting to .* cluster layer"
        ]
        watch = self.create_watch(watchpats, self._env["DeadTime"] + 10)
        watch.set_watch()

        self._cm.start_cm_async(node)
        ret = watch.look_for_all()
        if not ret:
            self._logger.log(f"Patterns not found: {watch.unmatched!r}")
            return self.failure(f"Setup of {node} failed")

        ret = self._stop(node)
        if not ret:
            return self.failure(f"{node} did not stop in time")

        return self.success()

    @property
    def errors_to_ignore(self):
        """Return a list of errors which should be ignored."""
        # We might do some fencing in the 2-node case if we make it up far enough
        return [
            r"Executing reboot fencing operation",
            r"Requesting fencing \([^)]+\) targeting node "
        ]



"""Corosync-specific class for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = ["Corosync2"]
__copyright__ = "Copyright 2007-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.CTS import Process
from pacemaker._cts.clustermanager import ClusterManager
from pacemaker._cts.patterns import PatternSelector

# Throughout this file, pylint has trouble understanding that EnvFactory
# is a singleton instance that can be treated as a subscriptable object.
# Various warnings are disabled because of this.  See also a comment about
# self._rsh in environment.py.
# pylint: disable=unsubscriptable-object


class Corosync2(ClusterManager):
    """A subclass of ClusterManager specialized to handle corosync2 and later based clusters."""

    def __init__(self):
        """Create a new Corosync2 instance."""
        ClusterManager.__init__(self)

        self._fullcomplist = {}
        self.templates = PatternSelector(self.name)

    @property
    def components(self):
        """Return a list of patterns that should be ignored for the cluster's components."""
        complist = []

        if not self._fullcomplist:
            common_ignore = self.templates.get_component("common-ignore")

            daemons = [
                "pacemaker-based",
                "pacemaker-controld",
                "pacemaker-attrd",
                "pacemaker-execd",
                "pacemaker-fenced"
            ]
            for c in daemons:
                badnews = self.templates.get_component(f"{c}-ignore") + common_ignore
                proc = Process(self, c, pats=self.templates.get_component(c),
                               badnews_ignore=badnews)
                self._fullcomplist[c] = proc

            # the scheduler uses dc_pats instead of pats
            badnews = self.templates.get_component("pacemaker-schedulerd-ignore") + common_ignore
            proc = Process(self, "pacemaker-schedulerd",
                           dc_pats=self.templates.get_component("pacemaker-schedulerd"),
                           badnews_ignore=badnews)
            self._fullcomplist["pacemaker-schedulerd"] = proc

            # add (or replace) extra components
            badnews = self.templates.get_component("corosync-ignore") + common_ignore
            proc = Process(self, "corosync", pats=self.templates.get_component("corosync"),
                           badnews_ignore=badnews)
            self._fullcomplist["corosync"] = proc

        # Processes running under valgrind can't be shot with "killall -9 processname",
        # so don't include them in the returned list
        vgrind = self.env["valgrind-procs"].split()
        for (key, val) in self._fullcomplist.items():
            if key in vgrind:
                self.log(f"Filtering {key} from the component list as it is being profiled by valgrind")
                continue

            if key == "pacemaker-fenced" and not self.env["DoFencing"]:
                continue

            complist.append(val)

        return complist



"""Log searching classes for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = ["LogKind", "LogWatcher"]
__copyright__ = "Copyright 2014-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from enum import Enum, auto, unique
import re
import time
import threading

from pacemaker.buildoptions import BuildOptions
from pacemaker._cts.errors import OutputNotFoundError
from pacemaker._cts.logging import LogFactory
from pacemaker._cts.remote import RemoteFactory

# dateutil >=2.7.0 provides isoparser, but we support older versions
try:
    from dateutil.parser import isoparser
except ImportError:
    import datetime
    import subprocess
    # pylint: disable=ungrouped-imports
    from pacemaker._cts.process import pipe_communicate

    # pylint: disable=invalid-name
    class isoparser():
        """Mimic isoparser.isoparse when not available."""

        def isoparse(self, timestamp):
            """Mimic isoparser.isoparse when not available."""
            with subprocess.Popen(["iso8601", "--output-as", "xml", "-d", timestamp],
                                  stdout=subprocess.PIPE) as result:

                output = pipe_communicate(result)
                if result.returncode != 0:
                    raise RuntimeError("Unable to run iso8601 command")
                match = re.search(r".*<date>\s*(\d+)-(\d+)-(\d+) (\d+):(\d+):(\d+)Z",
                                  output)
                if match is None:
                    raise RuntimeError("Unable to parse iso8601 command output")
                return datetime.datetime(int(match.group(1)),
                                         int(match.group(2)),
                                         int(match.group(3)),
                                         int(match.group(4)),
                                         int(match.group(5)),
                                         int(match.group(6)))


CTS_SUPPORT_BIN = f"{BuildOptions.DAEMON_DIR}/cts-support"


@unique
class LogKind(Enum):
    """The various kinds of log files that can be watched."""

    LOCAL_FILE = auto()     # From a local aggregation file on the exerciser
    REMOTE_FILE = auto()    # From a file on each cluster node
    JOURNAL = auto()        # From the systemd journal on each cluster node

    def __str__(self):
        """Return a printable string for a LogKind value."""
        return self.name.lower().replace('_', ' ')


class SearchObj:
    """
    The base class for various kinds of log watchers.

    Log-specific watchers need to be built on top of this one.
    """

    def __init__(self, filename, host=None, name=None):
        """
        Create a new SearchObj instance.

        Arguments:
        filename -- The log to watch
        host     -- The cluster node on which to watch the log
        name     -- A unique name to use when logging about this watch
        """
        self.filename = filename
        self.limit = None
        self.logger = LogFactory()
        self.name = name
        self.offset = "EOF"
        self.rsh = RemoteFactory().getInstance()

        if host:
            self.host = host
        else:
            self.host = "localhost"

        self._cache = []
        self._delegate = None

        async_task = self.harvest_async()
        async_task.join()

    def __str__(self):
        if self.host:
            return f"{self.host}:{self.filename}"

        return self.filename

    def log(self, args):
        """Log a message."""
        message = f"lw: {self}: {args}"
        self.logger.log(message)

    def debug(self, args):
        """Log a debug message."""
        message = f"lw: {self}: {args}"
        self.logger.debug(message)

    def harvest_async(self, delegate=None):
        """
        Collect lines from a log asynchronously.

        Optionally, also call delegate when complete.  This method must be
        implemented by all subclasses.
        """
        raise NotImplementedError

    def harvest_cached(self):
        """Return cached logs from before the limit timestamp."""
        raise NotImplementedError

    def end(self):
        """
        Mark that a log is done being watched.

        This function also resets internal data structures to the beginning
        of the file.  Subsequent watches will therefore start from the
        beginning again.
        """
        self.debug("Clearing cache and unsetting limit")
        self._cache = []
        self.limit = None


class FileObj(SearchObj):
    """A specialized SearchObj subclass for watching log files."""

    def __init__(self, filename, host=None, name=None):
        """
        Create a new FileObj instance.

        Arguments:
        filename -- The file to watch
        host     -- The cluster node on which to watch the file
        name     -- A unique name to use when logging about this watch
        """
        SearchObj.__init__(self, filename, host, name)

    def async_complete(self, pid, returncode, out, err):
        """
        Handle completion of an asynchronous log file read.

        This function saves the output from that read for look()/look_for_all()
        to process and records the current position in the journal.  Future
        reads will pick back up from that spot.

        Arguments:
        pid         -- The ID of the process that did the read
        returncode  -- The return code of the process that did the read
        out         -- stdout from the file read
        err         -- stderr from the file read
        """
        messages = []
        for line in out:
            match = re.search(r"^CTSwatcher:Last read: (\d+)", line)

            if match:
                self.offset = match.group(1)
                self.debug(f"Got {len(out)} lines, new offset: {self.offset}  {self._delegate!r}")
            elif re.search(r"^CTSwatcher:.*truncated", line):
                self.log(line)
            elif re.search(r"^CTSwatcher:", line):
                self.debug(f"Got control line: {line}")
            else:
                messages.append(line)

        if self._delegate:
            self._delegate.async_complete(pid, returncode, messages, err)

    def harvest_async(self, delegate=None):
        """
        Collect lines from the log file on a single host asynchronously.

        Optionally, call delegate when complete.  This can be called
        repeatedly, reading a chunk each time or until the end of the
        log file is hit.
        """
        self._delegate = delegate

        if self.limit and (self.offset == "EOF" or int(self.offset) > self.limit):
            if self._delegate:
                self._delegate.async_complete(-1, -1, [], [])

            return None

        cmd = f"{CTS_SUPPORT_BIN} watch -p CTSwatcher: -l 200 -f {self.filename} -o {self.offset}"
        return self.rsh.call_async(self.host, cmd, delegate=self)

    def harvest_cached(self):
        """Return cached logs from before the limit timestamp."""
        # cts-log-watcher script renders caching unnecessary for FileObj.
        # @TODO Caching might be slightly more efficient, if not too complex.
        return []

    def set_end(self):
        """
        Internally record where we expect to find the end of a log file.

        Calls to harvest from the log file will not go any farther than
        what this function records.
        """
        if self.limit:
            return

        cmd = f"{CTS_SUPPORT_BIN} watch -p CTSwatcher: -l 2 -f {self.filename} -o EOF"

        # pylint: disable=not-callable
        (_, lines) = self.rsh(self.host, cmd, verbose=0)

        for line in lines:
            match = re.search(r"^CTSwatcher:Last read: (\d+)", line)
            if match:
                self.limit = int(match.group(1))
                self.debug(f"Set limit to: {self.limit}")


class JournalObj(SearchObj):
    """A specialized SearchObj subclass for watching systemd journals."""

    def __init__(self, host=None, name=None):
        """
        Create a new JournalObj instance.

        Arguments:
        host     -- The cluster node on which to watch the journal
        name     -- A unique name to use when logging about this watch
        """
        SearchObj.__init__(self, "journal", host, name)
        self._parser = isoparser()

    def _msg_after_limit(self, msg):
        """
        Check whether a message was logged after the limit timestamp.

        Arguments:
        msg -- Message to check

        Returns `True` if `msg` was logged after `self.limit`, or `False`
        otherwise.
        """
        if not self.limit:
            return False

        match = re.search(r"^\S+", msg)
        if not match:
            return False

        msg_timestamp = match.group(0)
        msg_dt = self._parser.isoparse(msg_timestamp)
        return msg_dt > self.limit

    def _split_msgs_by_limit(self, msgs):
        """
        Split a sorted list of messages relative to the limit timestamp.

        Arguments:
        msgs -- List of messages to split

        Returns a tuple:
        (list of messages logged on or before limit timestamp,
         list of messages logged after limit timestamp).
        """
        # If last message was logged before limit, all messages were
        if msgs and self._msg_after_limit(msgs[-1]):

            # Else find index of first message logged after limit
            for idx, msg in enumerate(msgs):
                if self._msg_after_limit(msg):
                    self.debug(f"Got {idx} lines before passing limit timestamp")
                    return msgs[:idx], msgs[idx:]

        self.debug(f"Got {len(msgs)} lines")
        return msgs, []

    def async_complete(self, pid, returncode, out, err):
        """
        Handle completion of an asynchronous journal read.

        This function saves the output from that read for look()/look_for_all()
        to process and records the current position in the journal.  Future
        reads will pick back up from that spot.

        Arguments:
        pid         -- The ID of the process that did the journal read
        returncode  -- The return code of the process that did the journal read
        out         -- stdout from the journal read
        err         -- stderr from the journal read
        """
        if out:
            # Cursor should always be last line of journalctl output
            out, cursor_line = out[:-1], out[-1]
            match = re.search(r"^-- cursor: ([^.]+)", cursor_line)
            if not match:
                raise OutputNotFoundError(f"Cursor not found at end of output:\n{out}")

            self.offset = match.group(1).strip()
            self.debug(f"Got new cursor: {self.offset}")

        before, after = self._split_msgs_by_limit(out)

        # Save remaining messages after limit for later processing
        self._cache.extend(after)

        if self._delegate:
            self._delegate.async_complete(pid, returncode, before, err)

    def harvest_async(self, delegate=None):
        """
        Collect lines from the journal on a single host asynchronously.

        Optionally, call delegate when complete.  This can be called
        repeatedly, reading a chunk each time or until the end of the journal
        is hit.
        """
        self._delegate = delegate

        # Use --lines to prevent journalctl from overflowing the Popen input
        # buffer
        command = "journalctl --quiet --output=short-iso --show-cursor"
        if self.offset == "EOF":
            command += " --lines 0"
        else:
            command += f" --after-cursor='{self.offset}' --lines=200"

        return self.rsh.call_async(self.host, command, delegate=self)

    def harvest_cached(self):
        """Return cached logs from before the limit timestamp."""
        before, self._cache = self._split_msgs_by_limit(self._cache)
        return before

    def set_end(self):
        """
        Internally record where we expect to find the end of a host's journal.

        Calls to harvest from the journal will not go any farther than what
        this function records.
        """
        if self.limit:
            return

        # --iso-8601=seconds yields YYYY-MM-DDTHH:MM:SSZ, where Z is timezone
        # as offset from UTC

        # pylint: disable=not-callable
        (rc, lines) = self.rsh(self.host, "date --iso-8601=seconds", verbose=0)

        if rc == 0 and len(lines) == 1:
            self.limit = self._parser.isoparse(lines[0].strip())
            self.debug(f"Set limit to: {self.limit}")
        else:
            self.debug(f"Unable to set limit for {self.host} because date returned "
                       f"{len(lines)} lines with status {rc}")


class LogWatcher:
    """
    Watch a single log file or journal across multiple hosts.

    Instances of this class look for lines that match given regular
    expressions.

    The way you use this class is as follows:
        - Construct a LogWatcher object
        - Call set_watch() when you want to start watching the log
        - Call look() to scan the log looking for the patterns
    """

    def __init__(self, log, regexes, hosts, kind, name="Anon", timeout=10,
                 silent=False):
        """
        Create a new LogWatcher instance.

        Arguments:
        log     -- The log file to watch
        regexes -- A list of regular expressions to match against the log
        hosts   -- A list of cluster nodes on which to watch the log
        kind    -- What type of log is this object watching?
        name    -- A unique name to use when logging about this watch
        timeout -- Default number of seconds to watch a log file at a time;
                   this can be overridden by the timeout= parameter to
                   self.look on an as-needed basis
        silent  -- If False, log extra information
        """
        self.filename = log
        self.hosts = hosts
        self.kind = kind
        self.name = name
        self.regexes = regexes
        self.unmatched = None
        self.whichmatch = -1

        self._cache_lock = threading.Lock()
        self._file_list = []
        self._line_cache = []
        self._logger = LogFactory()
        self._timeout = int(timeout)

        #  Validate our arguments.  Better sooner than later ;-)
        for regex in regexes:
            re.compile(regex)

        if not self.hosts:
            raise ValueError("LogWatcher requires hosts argument")

        if self.kind != LogKind.JOURNAL and not self.filename:
            raise ValueError("LogWatcher requires log file name if not journal")

        if not silent:
            for regex in self.regexes:
                self._debug(f"Looking for regex: {regex}")

    def _debug(self, args):
        """Log a debug message."""
        message = f"lw: {self.name}: {args}"
        self._logger.debug(message)

    def set_watch(self):
        """Mark the place to start watching the log from."""
        if self.kind == LogKind.LOCAL_FILE:
            self._file_list.append(FileObj(self.filename))

        elif self.kind == LogKind.REMOTE_FILE:
            for node in self.hosts:
                self._file_list.append(FileObj(self.filename, node, self.name))

        elif self.kind == LogKind.JOURNAL:
            for node in self.hosts:
                self._file_list.append(JournalObj(node, self.name))

    def async_complete(self, pid, returncode, out, err):
        """
        Handle completion of an asynchronous log file read.

        This function saves the output from that read for look()/look_for_all()
        to process and records the current position.  Future reads will pick
        back up from that spot.

        Arguments:
        pid         -- The ID of the process that did the read
        returncode  -- The return code of the process that did the read
        out         -- stdout from the file read
        err         -- stderr from the file read
        """
        # Called as delegate through {File,Journal}Obj.async_complete()
        # pylint: disable=unused-argument

        # TODO: Probably need a lock for updating self._line_cache
        self._logger.debug(f"{self.name}: Got {len(out)} lines from {pid} (total {len(self._line_cache)})")

        if out:
            with self._cache_lock:
                self._line_cache.extend(out)

    def __get_lines(self):
        """Iterate over all watched log files and collect new lines from each."""
        if not self._file_list:
            raise ValueError("No sources to read from")

        pending = []

        for f in self._file_list:
            cached = f.harvest_cached()
            if cached:
                self._debug(f"Got {len(cached)} lines from {f.name} cache (total {len(self._line_cache)})")
                with self._cache_lock:
                    self._line_cache.extend(cached)
            else:
                t = f.harvest_async(self)
                if t:
                    pending.append(t)

        for t in pending:
            t.join(60.0)
            if t.is_alive():
                self._logger.log(f"{self.name}: Aborting after 20s waiting for {t!r} logging commands")
                return

    def end(self):
        """
        Mark that a log is done being watched.

        This function also resets internal data structures to the beginning
        of the file.  Subsequent watches will therefore start from the
        beginning again.
        """
        for f in self._file_list:
            f.end()

    def look(self, timeout=None):
        """
        Examine the log looking for the regexes in this object.

        It starts looking from the place marked by set_watch(), continuing
        through the file in the fashion of `tail -f`.  It properly recovers
        from log file truncation but not from removing and recreating the log.

        Arguments:
        timeout -- Number of seconds to watch the log file; defaults to
                   seconds argument passed when this object was created

        Returns the first line which matches any regex
        """
        if not timeout:
            timeout = self._timeout

        lines = 0
        begin = time.time()
        end = begin + timeout + 1

        if not self.regexes:
            self._debug("Nothing to look for")
            return None

        if timeout == 0:
            for f in self._file_list:
                f.set_end()

        while True:
            if self._line_cache:
                lines += 1

                with self._cache_lock:
                    line = self._line_cache[0]
                    self._line_cache.remove(line)

                which = -1

                if re.search("CTS:", line):
                    continue

                for regex in self.regexes:
                    which += 1

                    matchobj = re.search(regex, line)

                    if matchobj:
                        self.whichmatch = which
                        self._debug(f"Matched: {line}")
                        return line

            elif timeout > 0 and end < time.time():
                timeout = 0
                for f in self._file_list:
                    f.set_end()

            else:
                self.__get_lines()

                if not self._line_cache and end < time.time():
                    self._debug(f"Single search terminated: start={begin}, end={end}, now={time.time()}, lines={lines}")
                    return None

                self._debug(f"Waiting: start={begin}, end={end}, now={time.time()}, lines={len(self._line_cache)}")
                time.sleep(1)

    def look_for_all(self, allow_multiple_matches=False, silent=False):
        """
        Like look(), but looks for matches for multiple regexes.

        This function returns when the timeout is reached or all regexes were
        matched.  As a side effect, self.unmatched will contain regexes that
        were not matched.  This can be inspected by the caller.

        Arguments:
        allow_multiple_matches -- If True, allow each regex to match more than
                                  once.  If False (the default), once a regex
                                  matches a line, it will no longer be searched
                                  for.
        silent                 -- If False, log extra information

        Returns the matching lines if all regexes are matched, or None.
        """
        save_regexes = self.regexes
        result = []

        if not silent:
            self._debug(f"starting search: timeout={self._timeout}")

        while self.regexes:
            one_result = self.look(self._timeout)
            if not one_result:
                self.unmatched = self.regexes
                self.regexes = save_regexes
                self.end()
                return None

            result.append(one_result)
            if not allow_multiple_matches:
                del self.regexes[self.whichmatch]

            else:
                # Allow multiple regexes to match a single line
                tmp_regexes = self.regexes
                self.regexes = []

                for regex in tmp_regexes:
                    matchobj = re.search(regex, one_result)
                    if not matchobj:
                        self.regexes.append(regex)

        self.unmatched = None
        self.regexes = save_regexes
        return result



"""Restart the cluster and verify resources remain running."""

__all__ = ["Reattach"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import re
import time

from pacemaker.exitstatus import ExitStatus
from pacemaker._cts.audits import AuditResource
from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.simulstoplite import SimulStopLite
from pacemaker._cts.tests.starttest import StartTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._rsh is callable.
# pylint: disable=not-callable


class Reattach(CTSTest):
    """Restart the cluster and verify that resources remain running throughout."""

    def __init__(self, cm):
        """
        Create a new Reattach instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.name = "Reattach"

        self._startall = SimulStartLite(cm)
        self._stopall = SimulStopLite(cm)

    def _is_managed(self, node):
        """Return whether resources are managed by the cluster."""
        (_, is_managed) = self._rsh(node, "crm_attribute -t rsc_defaults -n is-managed -q -G -d true", verbose=1)
        is_managed = is_managed[0].strip()
        return is_managed == "true"

    def _set_unmanaged(self, node):
        """Disable resource management."""
        self.debug("Disable resource management")
        self._rsh(node, "crm_attribute -t rsc_defaults -n is-managed -v false")

    def _set_managed(self, node):
        """Enable resource management."""
        self.debug("Re-enable resource management")
        self._rsh(node, "crm_attribute -t rsc_defaults -n is-managed -D")

    def _disable_incompatible_rscs(self, node):
        """
        Disable resources that are incompatible with this test.

        Starts and stops of stonith-class resources are implemented internally
        by Pacemaker, which means that they must stop when Pacemaker is
        stopped, even if unmanaged. Disable them before running the Reattach
        test so they don't affect resource placement.

        Set target-role to "Stopped" for any of these resources in the CIB.
        """
        self.debug("Disable incompatible resources")
        xml = """'<meta_attributes id="cts-lab-Reattach-meta">
                    <nvpair id="cts-lab-Reattach-target-role" name="target-role" value="Stopped"/>
                    <rule id="cts-lab-Reattach-rule" boolean-op="or" score="INFINITY">
                      <rsc_expression id="cts-lab-Reattach-stonith" class="stonith"/>
                    </rule>
                  </meta_attributes>' --scope rsc_defaults"""
        return self._rsh(node, self._cm.templates['CibAddXml'] % xml)

    def _enable_incompatible_rscs(self, node):
        """Re-enable resources that were incompatible with this test."""
        self.debug("Re-enable incompatible resources")
        xml = """<meta_attributes id="cts-lab-Reattach-meta"/>"""
        return self._rsh(node, f"""cibadmin --delete --xml-text '{xml}'""")

    def _reprobe(self, node):
        """
        Reprobe all resources.

        The placement of some resources (such as promotable-1 in the
        lab-generated CIB) is affected by constraints using node-attribute-based
        rules. An earlier test may have erased the relevant node attribute, so
        do a reprobe, which should add the attribute back.
        """
        return self._rsh(node, """crm_resource --refresh""")

    def setup(self, node):
        """Set up this test."""
        if not self._startall(None):
            return self.failure("Startall failed")

        (rc, _) = self._disable_incompatible_rscs(node)
        if rc != ExitStatus.OK:
            return self.failure("Couldn't modify CIB to stop incompatible resources")

        (rc, _) = self._reprobe(node)
        if rc != ExitStatus.OK:
            return self.failure("Couldn't reprobe resources")

        if not self._cm.cluster_stable(double_check=True):
            return self.failure("Cluster did not stabilize after setup")

        return self.success()

    def teardown(self, node):
        """Tear down this test."""
        # Make sure 'node' is up
        start = StartTest(self._cm)
        start(node)

        if not self._is_managed(node):
            self._set_managed(node)

        (rc, _) = self._enable_incompatible_rscs(node)
        if rc != ExitStatus.OK:
            return self.failure("Couldn't modify CIB to re-enable incompatible resources")

        if not self._cm.cluster_stable():
            return self.failure("Cluster did not stabilize after teardown")
        if not self._is_managed(node):
            return self.failure("Could not re-enable resource management")

        return self.success()

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")

        # Conveniently, the scheduler will display this message when disabling
        # management, even if fencing is not enabled, so we can rely on it.
        managed = self.create_watch(["No fencing will be done"], 60)
        managed.set_watch()

        self._set_unmanaged(node)

        if not managed.look_for_all():
            self._logger.log(f"Patterns not found: {managed.unmatched!r}")
            return self.failure("Resource management not disabled")

        pats = [
            self.templates["Pat:RscOpOK"] % ("start", ".*"),
            self.templates["Pat:RscOpOK"] % ("stop", ".*"),
            self.templates["Pat:RscOpOK"] % ("promote", ".*"),
            self.templates["Pat:RscOpOK"] % ("demote", ".*"),
            self.templates["Pat:RscOpOK"] % ("migrate", ".*")
        ]

        watch = self.create_watch(pats, 60, "ShutdownActivity")
        watch.set_watch()

        self.debug("Shutting down the cluster")
        ret = self._stopall(None)
        if not ret:
            self._set_managed(node)
            return self.failure("Couldn't shut down the cluster")

        self.debug("Bringing the cluster back up")
        ret = self._startall(None)
        time.sleep(5)  # allow ping to update the CIB
        if not ret:
            self._set_managed(node)
            return self.failure("Couldn't restart the cluster")

        if self.local_badnews("ResourceActivity:", watch):
            self._set_managed(node)
            return self.failure("Resources stopped or started during cluster restart")

        watch = self.create_watch(pats, 60, "StartupActivity")
        watch.set_watch()

        # Re-enable resource management (and verify it happened).
        self._set_managed(node)
        self._cm.cluster_stable()
        if not self._is_managed(node):
            return self.failure("Could not re-enable resource management")

        # Ignore actions for STONITH resources
        ignore = []
        (_, lines) = self._rsh(node, "crm_resource -c", verbose=1)
        for line in lines:
            if re.search("^Resource", line):
                r = AuditResource(self._cm, line)

                if r.rclass == "stonith":
                    self.debug(f"Ignoring start actions for {r.id}")
                    ignore.append(self.templates["Pat:RscOpOK"] % ("start", r.id))

        if self.local_badnews("ResourceActivity:", watch, ignore):
            return self.failure("Resources stopped or started after resource management was re-enabled")

        return ret

    @property
    def errors_to_ignore(self):
        """Return a list of errors which should be ignored."""
        return [
            r"resource( was|s were) active at shutdown"
        ]



"""Stop running nodes, and start stopped nodes."""

__all__ = ["FlipTest"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import time

from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.starttest import StartTest
from pacemaker._cts.tests.stoptest import StopTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class FlipTest(CTSTest):
    """Stop running nodes and start stopped nodes."""

    def __init__(self, cm):
        """
        Create a new FlipTest instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)
        self.name = "Flip"

        self._start = StartTest(cm)
        self._stop = StopTest(cm)

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")

        if self._cm.expected_status[node] == "up":
            self.incr("stopped")
            ret = self._stop(node)
            kind = "up->down"
            # Give the cluster time to recognize it's gone...
            time.sleep(self._env["StableTime"])
        elif self._cm.expected_status[node] == "down":
            self.incr("started")
            ret = self._start(node)
            kind = "down->up"
        else:
            return self.skipped()

        self.incr(kind)
        if ret:
            return self.success()

        return self.failure(f"{kind} failure")



"""Test classes for the `pacemaker._cts` package."""

__copyright__ = "Copyright 2023-2024 the Pacemaker project contributors"
__license__ = "GNU Lesser General Public License version 2.1 or later (LGPLv2.1+)"

from pacemaker._cts.tests.componentfail import ComponentFail
from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.fliptest import FlipTest
from pacemaker._cts.tests.maintenancemode import MaintenanceMode
from pacemaker._cts.tests.nearquorumpointtest import NearQuorumPointTest
from pacemaker._cts.tests.partialstart import PartialStart
from pacemaker._cts.tests.reattach import Reattach
from pacemaker._cts.tests.restartonebyone import RestartOnebyOne
from pacemaker._cts.tests.resourcerecover import ResourceRecover
from pacemaker._cts.tests.restarttest import RestartTest
from pacemaker._cts.tests.resynccib import ResyncCIB
from pacemaker._cts.tests.remotebasic import RemoteBasic
from pacemaker._cts.tests.remotedriver import RemoteDriver
from pacemaker._cts.tests.remotemigrate import RemoteMigrate
from pacemaker._cts.tests.remoterscfailure import RemoteRscFailure
from pacemaker._cts.tests.remotestonithd import RemoteStonithd
from pacemaker._cts.tests.simulstart import SimulStart
from pacemaker._cts.tests.simulstop import SimulStop
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.simulstoplite import SimulStopLite
from pacemaker._cts.tests.splitbraintest import SplitBrainTest
from pacemaker._cts.tests.standbytest import StandbyTest
from pacemaker._cts.tests.starttest import StartTest
from pacemaker._cts.tests.startonebyone import StartOnebyOne
from pacemaker._cts.tests.stonithdtest import StonithdTest
from pacemaker._cts.tests.stoponebyone import StopOnebyOne
from pacemaker._cts.tests.stoptest import StopTest


def test_list(cm, audits):
    """
    Return a list of runnable test class objects.

    These are objects that are enabled and whose is_applicable methods return
    True.
    """
    # cm is a reasonable name here.
    # pylint: disable=invalid-name

    # A list of all enabled test classes, in the order that they should
    # be run (if we're doing --once).  There are various other ways of
    # specifying which tests should be run, in which case the order here
    # will not matter.
    #
    # Note that just because a test is listed here doesn't mean it will
    # definitely be run - is_applicable is still taken into consideration.
    # Also note that there are other tests that are excluded from this
    # list for various reasons.
    enabled_test_classes = [
        FlipTest,
        RestartTest,
        StonithdTest,
        StartOnebyOne,
        SimulStart,
        SimulStop,
        StopOnebyOne,
        RestartOnebyOne,
        PartialStart,
        StandbyTest,
        MaintenanceMode,
        ResourceRecover,
        ComponentFail,
        SplitBrainTest,
        Reattach,
        ResyncCIB,
        NearQuorumPointTest,
        RemoteBasic,
        RemoteStonithd,
        RemoteMigrate,
        RemoteRscFailure,
    ]

    result = []

    for testclass in enabled_test_classes:
        bound_test = testclass(cm)

        if bound_test.is_applicable():
            bound_test.audits = audits
            result.append(bound_test)

    return result



"""Stop all running nodes simultaneously."""

__all__ = ["SimulStop"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.simulstoplite import SimulStopLite


class SimulStop(CTSTest):
    """Stop all running nodes simultaneously."""

    def __init__(self, cm):
        """
        Create a new SimulStop instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.name = "SimulStop"

        self._startall = SimulStartLite(cm)
        self._stopall = SimulStopLite(cm)

    def __call__(self, dummy):
        """Perform this test."""
        self.incr("calls")

        ret = self._startall(None)
        if not ret:
            return self.failure("Setup failed")

        if not self._stopall(None):
            return self.failure("Stopall failed")

        return self.success()



"""A module providing private library management code."""

__all__ = ["_libcrmcommon"]
__copyright__ = "Copyright 2024-2025 the Pacemaker project contributors"
__license__ = "GNU Lesser General Public License version 2.1 or later (LGPLv2.1+)"

import ctypes
from ctypes.util import find_library
from glob import glob
import os

from pacemaker.buildoptions import BuildOptions


def load_library(basename):
    """Find and load the library with the given base name."""
    path = find_library(basename)

    # If the library was not found anywhere in the default locations, also search
    # for it in the build directory
    if path is None:
        # pylint: disable=protected-access
        for d in glob(f"{BuildOptions._BUILD_DIR}/lib/*/.libs"):
            path = f"{d}/lib{basename}.so"

            if os.path.exists(path):
                break

            path = None

    if path is None:
        raise FileNotFoundError(basename)

    return ctypes.cdll.LoadLibrary(path)


_libcrmcommon = load_library("crmcommon")
_libcrmcommon.crm_exit_str.restype = ctypes.c_char_p



"""Test environment classes for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = ["EnvFactory"]
__copyright__ = "Copyright 2014-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import argparse
from contextlib import suppress
import os
import random
import socket
import sys
import time

from pacemaker.buildoptions import BuildOptions
from pacemaker._cts.logging import LogFactory
from pacemaker._cts.remote import RemoteFactory
from pacemaker._cts.watcher import LogKind


class Environment:
    """
    A class for managing the CTS environment.

    This consists largely of processing and storing command line parameters.
    """

    # pylint doesn't understand that self._rsh is callable (it stores the
    # singleton instance of RemoteExec, as returned by the getInstance method
    # of RemoteFactory).
    # @TODO See if type annotations fix this.

    # I think we could also fix this by getting rid of the getInstance methods,
    # but that's a project for another day.  For now, just disable the warning.
    # pylint: disable=not-callable

    def __init__(self, args):
        """
        Create a new Environment instance.

        This class can be treated kind of like a dictionary due to the presence
        of typical dict functions like __contains__, __getitem__, and __setitem__.
        However, it is not a dictionary so do not rely on standard dictionary
        behavior.

        Arguments:
        args -- A list of command line parameters, minus the program name.
                If None, sys.argv will be used.
        """
        self.data = {}
        self._nodes = []

        # Set some defaults before processing command line arguments.  These are
        # either not set by any command line parameter, or they need a default
        # that can't be set in add_argument.
        self["DeadTime"] = 300
        self["StartTime"] = 300
        self["StableTime"] = 30
        self["tests"] = []
        self["IPagent"] = "IPaddr2"
        self["DoFencing"] = True
        self["ClobberCIB"] = False
        self["CIBfilename"] = None
        self["CIBResource"] = False
        self["log_kind"] = None
        self["node-limit"] = 0
        self["scenario"] = "random"

        self.random_gen = random.Random()

        self._logger = LogFactory()
        self._rsh = RemoteFactory().getInstance()
        self._target = "localhost"

        self._seed_random()
        self._parse_args(args)

        if not self["ListTests"]:
            self._validate()
            self._discover()

    def _seed_random(self, seed=None):
        """
        Initialize the random number generator.

        Arguments:
        seed -- Use this to see the random number generator, or use the
                current time if None.
        """
        if not seed:
            seed = int(time.time())

        self["RandSeed"] = seed
        self.random_gen.seed(str(seed))

    def dump(self):
        """Print the current environment."""
        keys = []
        for key in list(self.data.keys()):
            keys.append(key)

        keys.sort()
        for key in keys:
            self._logger.debug(f"{f'Environment[{key}]':35}: {str(self[key])}")

    def keys(self):
        """Return a list of all environment keys stored in this instance."""
        return list(self.data.keys())

    def __contains__(self, key):
        """Return True if the given key exists in the environment."""
        if key == "nodes":
            return True

        return key in self.data

    def __getitem__(self, key):
        """Return the given environment key, or None if it does not exist."""
        if str(key) == "0":
            raise ValueError("Bad call to 'foo in X', should reference 'foo in X.keys()' instead")

        if key == "nodes":
            return self._nodes

        if key == "Name":
            return self._get_stack_short()

        return self.data.get(key)

    def __setitem__(self, key, value):
        """Set the given environment key to the given value, overriding any previous value."""
        if key == "Stack":
            self._set_stack(value)

        elif key == "node-limit":
            self.data[key] = value
            self._filter_nodes()

        elif key == "nodes":
            self._nodes = []
            for node in value:
                # I don't think I need the IP address, etc. but this validates
                # the node name against /etc/hosts and/or DNS, so it's a
                # GoodThing(tm).
                try:
                    n = node.strip()
                    # @TODO This only handles IPv4, use getaddrinfo() instead
                    # (here and in _discover())
                    socket.gethostbyname_ex(n)
                    self._nodes.append(n)
                except socket.herror:
                    self._logger.log(f"{node} not found in DNS... aborting")
                    raise

            self._filter_nodes()

        else:
            self.data[key] = value

    def random_node(self):
        """Choose a random node from the cluster."""
        return self.random_gen.choice(self["nodes"])

    def get(self, key, default=None):
        """Return the value for key if key is in the environment, else default."""
        if key == "nodes":
            return self._nodes

        return self.data.get(key, default)

    def _set_stack(self, name):
        """Normalize the given cluster stack name."""
        if name in ["corosync", "cs", "mcp"]:
            self.data["Stack"] = "corosync 2+"

        else:
            raise ValueError(f"Unknown stack: {name}")

    def _get_stack_short(self):
        """Return the short name for the currently set cluster stack."""
        if "Stack" not in self.data:
            return "unknown"

        if self.data["Stack"] == "corosync 2+":
            return "crm-corosync"

        LogFactory().log(f"Unknown stack: {self['stack']}")
        raise ValueError(f"Unknown stack: {self['stack']}")

    def _detect_systemd(self):
        """Detect whether systemd is in use on the target node."""
        if "have_systemd" not in self.data:
            (rc, _) = self._rsh(self._target, "systemctl list-units", verbose=0)
            self["have_systemd"] = rc == 0

    def _detect_syslog(self):
        """Detect the syslog variant in use on the target node (if any)."""
        if "syslogd" in self.data:
            return

        if self["have_systemd"]:
            # Systemd
            (_, lines) = self._rsh(self._target, r"systemctl list-units | grep syslog.*\.service.*active.*running | sed 's:.service.*::'", verbose=1)
        else:
            # SYS-V
            (_, lines) = self._rsh(self._target, "chkconfig --list | grep syslog.*on | awk '{print $1}' | head -n 1", verbose=1)

        with suppress(IndexError):
            self["syslogd"] = lines[0].strip()

    def disable_service(self, node, service):
        """Disable the given service on the given node."""
        if self["have_systemd"]:
            # Systemd
            (rc, _) = self._rsh(node, f"systemctl disable {service}")
            return rc

        # SYS-V
        (rc, _) = self._rsh(node, f"chkconfig {service} off")
        return rc

    def enable_service(self, node, service):
        """Enable the given service on the given node."""
        if self["have_systemd"]:
            # Systemd
            (rc, _) = self._rsh(node, f"systemctl enable {service}")
            return rc

        # SYS-V
        (rc, _) = self._rsh(node, f"chkconfig {service} on")
        return rc

    def service_is_enabled(self, node, service):
        """Return True if the given service is enabled on the given node."""
        if self["have_systemd"]:
            # Systemd

            # With "systemctl is-enabled", we should check if the service is
            # explicitly "enabled" instead of the return code. For example it returns
            # 0 if the service is "static" or "indirect", but they don't really count
            # as "enabled".
            (rc, _) = self._rsh(node, f"systemctl is-enabled {service} | grep enabled")
            return rc == 0

        # SYS-V
        (rc, _) = self._rsh(node, f"chkconfig --list | grep -e {service}.*on")
        return rc == 0

    def _detect_at_boot(self):
        """Detect if the cluster starts at boot."""
        if "at-boot" not in self.data:
            self["at-boot"] = self.service_is_enabled(self._target, "corosync") \
                or self.service_is_enabled(self._target, "pacemaker")

    def _detect_ip_offset(self):
        """Detect the offset for IPaddr resources."""
        if self["CIBResource"] and "IPBase" not in self.data:
            (_, lines) = self._rsh(self._target, "ip addr | grep inet | grep -v -e link -e inet6 -e '/32' -e ' lo' | awk '{print $2}'", verbose=0)
            network = lines[0].strip()

            (_, lines) = self._rsh(self._target, "nmap -sn -n %s | grep 'scan report' | awk '{print $NF}' | sed 's:(::' | sed 's:)::' | sort -V | tail -n 1" % network, verbose=0)

            try:
                self["IPBase"] = lines[0].strip()
            except (IndexError, TypeError):
                self["IPBase"] = None

            if not self["IPBase"]:
                self["IPBase"] = " fe80::1234:56:7890:1000"
                self._logger.log("Could not determine an offset for IPaddr resources.  Perhaps nmap is not installed on the nodes.")
                self._logger.log(f"""Defaulting to '{self["IPBase"]}', use --test-ip-base to override""")
                return

            # pylint thinks self["IPBase"] is a list, not a string, which causes it
            # to error out because a list doesn't have split().
            # pylint: disable=no-member
            last_part = self["IPBase"].split('.')[3]

            if int(last_part) >= 240:
                self._logger.log(f"Could not determine an offset for IPaddr resources. Upper bound is too high: {self['IPBase']} {last_part}")
                self["IPBase"] = " fe80::1234:56:7890:1000"
                self._logger.log(f"""Defaulting to '{self["IPBase"]}', use --test-ip-base to override""")

    def _filter_nodes(self):
        """
        Filter the list of cluster nodes.

        If --limit-nodes is given, keep that many nodes from the front of the
        list of cluster nodes and drop the rest.
        """
        if self["node-limit"] > 0:
            if len(self["nodes"]) > self["node-limit"]:
                self._logger.log(f"Limiting the number of nodes configured={len(self['nodes'])} "
                                 f"(max={self['node-limit']})")

                while len(self["nodes"]) > self["node-limit"]:
                    self["nodes"].pop(len(self["nodes"]) - 1)

    def _validate(self):
        """Check that we were given all required command line parameters."""
        if not self["nodes"]:
            raise ValueError("No nodes specified!")

    def _discover(self):
        """Probe cluster nodes to figure out how to log and manage services."""
        self._target = random.Random().choice(self["nodes"])

        exerciser = socket.gethostname()

        # Use the IP where possible to avoid name lookup failures
        for ip in socket.gethostbyname_ex(exerciser)[2]:
            if ip != "127.0.0.1":
                exerciser = ip
                break

        self["cts-exerciser"] = exerciser

        self._detect_systemd()
        self._detect_syslog()
        self._detect_at_boot()
        self._detect_ip_offset()

    def _parse_args(self, argv):
        """
        Parse and validate command line parameters.

        Set the appropriate values in the environment dictionary.  If argv is
        None, use sys.argv instead.
        """
        if not argv:
            argv = sys.argv[1:]

        parser = argparse.ArgumentParser(epilog=f"{sys.argv[0]} -g virt1 -r --stonith ssh --schema pacemaker-2.0 500")

        grp1 = parser.add_argument_group("Common options")
        grp1.add_argument("-g", "--dsh-group", "--group",
                          metavar="GROUP", dest="group",
                          help="Use the nodes listed in the named DSH group (~/.dsh/groups/$name)")
        grp1.add_argument("-l", "--limit-nodes",
                          type=int, default=0,
                          metavar="MAX",
                          help="Only use the first MAX cluster nodes supplied with --nodes")
        grp1.add_argument("--benchmark",
                          action="store_true",
                          help="Add timing information")
        grp1.add_argument("--list", "--list-tests",
                          action="store_true", dest="list_tests",
                          help="List the valid tests")
        grp1.add_argument("--nodes",
                          metavar="NODES",
                          help="List of cluster nodes separated by whitespace")
        grp1.add_argument("--stack",
                          default="corosync",
                          metavar="STACK",
                          help="Which cluster stack is installed")

        grp2 = parser.add_argument_group("Options that CTS will usually auto-detect correctly")
        grp2.add_argument("-L", "--logfile",
                          metavar="PATH",
                          help="Where to look for logs from cluster nodes (or 'journal' for systemd journal)")
        grp2.add_argument("--at-boot", "--cluster-starts-at-boot",
                          choices=["1", "0", "yes", "no"],
                          help="Does the cluster software start at boot time?")
        grp2.add_argument("--facility", "--syslog-facility",
                          default="daemon",
                          metavar="NAME",
                          help="Which syslog facility to log to")
        grp2.add_argument("--ip", "--test-ip-base",
                          metavar="IP",
                          help="Offset for generated IP address resources")

        grp3 = parser.add_argument_group("Options for release testing")
        grp3.add_argument("-r", "--populate-resources",
                          action="store_true",
                          help="Generate a sample configuration")
        grp3.add_argument("--choose",
                          metavar="NAME",
                          help="Run only the named tests, separated by whitespace")
        grp3.add_argument("--fencing", "--stonith",
                          choices=["1", "0", "yes", "no", "lha", "openstack", "rhcs", "rhevm", "scsi", "ssh", "virt", "xvm"],
                          default="1",
                          help="What fencing agent to use")
        grp3.add_argument("--once",
                          action="store_true",
                          help="Run all valid tests once")

        grp4 = parser.add_argument_group("Additional (less common) options")
        grp4.add_argument("-c", "--clobber-cib",
                          action="store_true",
                          help="Erase any existing configuration")
        grp4.add_argument("-y", "--yes",
                          action="store_true", dest="always_continue",
                          help="Continue to run whenever prompted")
        grp4.add_argument("--boot",
                          action="store_true",
                          help="")
        grp4.add_argument("--cib-filename",
                          metavar="PATH",
                          help="Install the given CIB file to the cluster")
        grp4.add_argument("--experimental-tests",
                          action="store_true",
                          help="Include experimental tests")
        grp4.add_argument("--loop-minutes",
                          type=int, default=60,
                          help="")
        grp4.add_argument("--no-loop-tests",
                          action="store_true",
                          help="Don't run looping/time-based tests")
        grp4.add_argument("--no-unsafe-tests",
                          action="store_true",
                          help="Don't run tests that are unsafe for use with ocfs2/drbd")
        grp4.add_argument("--notification-agent",
                          metavar="PATH",
                          default="/var/lib/pacemaker/notify.sh",
                          help="Script to configure for Pacemaker alerts")
        grp4.add_argument("--notification-recipient",
                          metavar="R",
                          default="/var/lib/pacemaker/notify.log",
                          help="Recipient to pass to alert script")
        grp4.add_argument("--oprofile",
                          metavar="NODES",
                          help="List of cluster nodes to run oprofile on")
        grp4.add_argument("--outputfile",
                          metavar="PATH",
                          help="Location to write logs to")
        grp4.add_argument("--qarsh",
                          action="store_true",
                          help="Use QARSH to access nodes instead of SSH")
        grp4.add_argument("--schema",
                          metavar="SCHEMA",
                          default=f"pacemaker-{BuildOptions.CIB_SCHEMA_VERSION}",
                          help="Create a CIB conforming to the given schema")
        grp4.add_argument("--seed",
                          metavar="SEED",
                          help="Use the given string as the random number seed")
        grp4.add_argument("--set",
                          action="append",
                          metavar="ARG",
                          default=[],
                          help="Set key=value pairs (can be specified multiple times)")
        grp4.add_argument("--stonith-args",
                          metavar="ARGS",
                          default="hostlist=all,livedangerously=yes",
                          help="")
        grp4.add_argument("--stonith-type",
                          metavar="TYPE",
                          default="external/ssh",
                          help="")
        grp4.add_argument("--trunc",
                          action="store_true", dest="truncate",
                          help="Truncate log file before starting")
        grp4.add_argument("--valgrind-procs",
                          metavar="PROCS",
                          default="pacemaker-attrd pacemaker-based pacemaker-controld pacemaker-execd pacemaker-fenced pacemaker-schedulerd",
                          help="Run valgrind against the given space-separated list of processes")
        grp4.add_argument("--warn-inactive",
                          action="store_true",
                          help="Warn if a resource is assigned to an inactive node")

        parser.add_argument("iterations",
                            nargs='?',
                            type=int, default=1,
                            help="Number of tests to run")

        args = parser.parse_args(args=argv)

        # Set values on this object based on what happened with command line
        # processing.  This has to be done in several blocks.

        # These values can always be set.  They get a default from the add_argument
        # calls, only do one thing, and they do not have any side effects.
        self["ClobberCIB"] = args.clobber_cib
        self["ListTests"] = args.list_tests
        self["Schema"] = args.schema
        self["Stack"] = args.stack
        self["SyslogFacility"] = args.facility
        self["TruncateLog"] = args.truncate
        self["at-boot"] = args.at_boot in ["1", "yes"]
        self["benchmark"] = args.benchmark
        self["continue"] = args.always_continue
        self["experimental-tests"] = args.experimental_tests
        self["iterations"] = args.iterations
        self["loop-minutes"] = args.loop_minutes
        self["loop-tests"] = not args.no_loop_tests
        self["notification-agent"] = args.notification_agent
        self["notification-recipient"] = args.notification_recipient
        self["node-limit"] = args.limit_nodes
        self["stonith-params"] = args.stonith_args
        self["stonith-type"] = args.stonith_type
        self["unsafe-tests"] = not args.no_unsafe_tests
        self["valgrind-procs"] = args.valgrind_procs
        self["warn-inactive"] = args.warn_inactive

        # Nodes and groups are mutually exclusive, so their defaults cannot be
        # set in their add_argument calls.  Additionally, groups does more than
        # just set a value.  Here, set nodes first and then if a group is
        # specified, override the previous nodes value.
        if args.nodes:
            self["nodes"] = args.nodes.split(" ")
        else:
            self["nodes"] = []

        if args.group:
            self["OutputFile"] = f"{os.environ['HOME']}/cluster-{args.dsh_group}.log"
            LogFactory().add_file(self["OutputFile"], "CTS")

            dsh_file = f"{os.environ['HOME']}/.dsh/group/{args.dsh_group}"

            if os.path.isfile(dsh_file):
                self["nodes"] = []

                with open(dsh_file, "r", encoding="utf-8") as f:
                    for line in f:
                        stripped = line.strip()

                        if not stripped.startswith('#'):
                            self["nodes"].append(stripped)
            else:
                print(f"Unknown DSH group: {args.dsh_group}")

        # Everything else either can't have a default set in an add_argument
        # call (likely because we don't want to always have a value set for it)
        # or it does something fancier than just set a single value.  However,
        # order does not matter for these as long as the user doesn't provide
        # conflicting arguments on the command line.  So just do Everything
        # alphabetically.
        if args.boot:
            self["scenario"] = "boot"

        if args.cib_filename:
            self["CIBfilename"] = args.cib_filename
        else:
            self["CIBfilename"] = None

        if args.choose:
            self["scenario"] = "sequence"
            self["tests"].extend(args.choose.split())
            self["iterations"] = len(self["tests"])

        if args.fencing:
            if args.fencing in ["0", "no"]:
                self["DoFencing"] = False
            else:
                self["DoFencing"] = True

                if args.fencing in ["rhcs", "virt", "xvm"]:
                    self["stonith-type"] = "fence_xvm"

                elif args.fencing == "scsi":
                    self["stonith-type"] = "fence_scsi"

                elif args.fencing in ["lha", "ssh"]:
                    self["stonith-params"] = "hostlist=all,livedangerously=yes"
                    self["stonith-type"] = "external/ssh"

                elif args.fencing == "openstack":
                    self["stonith-type"] = "fence_openstack"

                    print("Obtaining OpenStack credentials from the current environment")
                    region = os.environ['OS_REGION_NAME']
                    tenant = os.environ['OS_TENANT_NAME']
                    auth = os.environ['OS_AUTH_URL']
                    user = os.environ['OS_USERNAME']
                    password = os.environ['OS_PASSWORD']

                    self["stonith-params"] = f"region={region},tenant={tenant},auth={auth},user={user},password={password}"

                elif args.fencing == "rhevm":
                    self["stonith-type"] = "fence_rhevm"

                    print("Obtaining RHEV-M credentials from the current environment")
                    user = os.environ['RHEVM_USERNAME']
                    password = os.environ['RHEVM_PASSWORD']
                    server = os.environ['RHEVM_SERVER']
                    port = os.environ['RHEVM_PORT']

                    self["stonith-params"] = f"login={user},passwd={password},ipaddr={server},ipport={port},ssl=1,shell_timeout=10"

        if args.ip:
            self["CIBResource"] = True
            self["ClobberCIB"] = True
            self["IPBase"] = args.ip

        if args.logfile == "journal":
            self["LogAuditDisabled"] = True
            self["log_kind"] = LogKind.JOURNAL
        elif args.logfile:
            self["LogAuditDisabled"] = True
            self["LogFileName"] = args.logfile
            self["log_kind"] = LogKind.REMOTE_FILE
        else:
            # We can't set this as the default on the parser.add_argument call
            # for this option because then args.logfile will be set, which means
            # the above branch will be taken and those other values will also be
            # set.
            self["LogFileName"] = "/var/log/messages"

        if args.once:
            self["scenario"] = "all-once"

        if args.oprofile:
            self["oprofile"] = args.oprofile.split(" ")
        else:
            self["oprofile"] = []

        if args.outputfile:
            self["OutputFile"] = args.outputfile
            LogFactory().add_file(self["OutputFile"])

        if args.populate_resources:
            self["CIBResource"] = True
            self["ClobberCIB"] = True

        if args.qarsh:
            self._rsh.enable_qarsh()

        for kv in args.set:
            (name, value) = kv.split("=")
            self[name] = value
            print(f"Setting {name} = {value}")


class EnvFactory:
    """A class for constructing a singleton instance of an Environment object."""

    instance = None

    # pylint: disable=invalid-name
    def getInstance(self, args=None):
        """
        Return the previously created instance of Environment.

        If no instance exists, create a new instance and return that.
        """
        if not EnvFactory.instance:
            EnvFactory.instance = Environment(args)

        return EnvFactory.instance



"""Start the cluster manager on a given node."""

__all__ = ["StartTest"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class StartTest(CTSTest):
    """
    A pseudo-test that sets up conditions before running some other test.

    This class starts the cluster manager on a given node.  Other test classes
    should not use this one as a superclass.
    """

    def __init__(self, cm):
        """
        Create a new StartTest instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)
        self.name = "Start"

    def __call__(self, node):
        """Start the given node, returning whether this succeeded or not."""
        self.incr("calls")

        if self._cm.upcount() == 0:
            self.incr("us")
        else:
            self.incr("them")

        if self._cm.expected_status[node] != "down":
            return self.skipped()

        if self._cm.start_cm(node):
            return self.success()

        return self.failure(f"Startup {self._env['Name']} on node {node} failed")



"""Base classes for CTS tests."""

__all__ = ["RemoteDriver"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import os
import time
import subprocess
import tempfile

from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.starttest import StartTest
from pacemaker._cts.tests.stoptest import StopTest
from pacemaker._cts.timer import Timer

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._rsh is callable.
# pylint: disable=not-callable


class RemoteDriver(CTSTest):
    """
    A specialized base class for cluster tests that run on Pacemaker Remote nodes.

    This builds on top of CTSTest to provide methods for starting and stopping
    services and resources, and managing remote nodes.  This is still just an
    abstract class -- specific tests need to implement their own specialized
    behavior.
    """

    def __init__(self, cm):
        """
        Create a new RemoteDriver instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)
        self.name = "RemoteDriver"

        self._corosync_enabled = False
        self._pacemaker_enabled = False
        self._remote_node = None
        self._remote_rsc = "remote-rsc"
        self._start = StartTest(cm)
        self._startall = SimulStartLite(cm)
        self._stop = StopTest(cm)

        self.reset()

    def reset(self):
        """Reset the state of this test back to what it was before the test was run."""
        self.failed = False
        self.fail_string = ""

        self._pcmk_started = False
        self._remote_node_added = False
        self._remote_rsc_added = False
        self._remote_use_reconnect_interval = self._env.random_gen.choice([True, False])

    def fail(self, msg):
        """Mark test as failed."""
        self.failed = True

        # Always log the failure.
        self._logger.log(msg)

        # Use first failure as test status, as it's likely to be most useful.
        if not self.fail_string:
            self.fail_string = msg

    def _get_other_node(self, node):
        """
        Get the first cluster node out of the environment that is not the given node.

        Typically, this is used to find some node that will still be active that
        we can run cluster commands on.
        """
        for othernode in self._env["nodes"]:
            if othernode == node:
                # we don't want to try and use the cib that we just shutdown.
                # find a cluster node that is not our soon to be remote-node.
                continue

            return othernode

    def _del_rsc(self, node, rsc):
        """
        Delete the given named resource from the cluster.

        The given `node` is the cluster node on which we should *not* run the
        delete command.
        """
        othernode = self._get_other_node(node)
        (rc, _) = self._rsh(othernode, f"crm_resource -D -r {rsc} -t primitive")
        if rc != 0:
            self.fail(f"Removal of resource '{rsc}' failed")

    def _add_rsc(self, node, rsc_xml):
        """
        Add a resource given in XML format to the cluster.

        The given `node` is the cluster node on which we should *not* run the
        add command.
        """
        othernode = self._get_other_node(node)
        (rc, _) = self._rsh(othernode, f"cibadmin -C -o resources -X '{rsc_xml}'")
        if rc != 0:
            self.fail("resource creation failed")

    def _add_primitive_rsc(self, node):
        """
        Add a primitive heartbeat resource for the remote node to the cluster.

        The given `node` is the cluster node on which we should *not* run the
        add command.
        """
        rsc_xml = f"""
<primitive class="ocf" id="{self._remote_rsc}" provider="heartbeat" type="Dummy">
  <meta_attributes id="{self._remote_rsc}-meta_attributes"/>
  <operations>
    <op id="{self._remote_rsc}-monitor-interval-20s" interval="20s" name="monitor"/>
  </operations>
</primitive>"""

        self._add_rsc(node, rsc_xml)
        if not self.failed:
            self._remote_rsc_added = True

    def _add_connection_rsc(self, node):
        """
        Add a primitive connection resource for the remote node to the cluster.

        The given `node` is the cluster node on which we should *not* run the
        add command.
        """
        rsc_xml = f"""
<primitive class="ocf" id="{self._remote_node}" provider="pacemaker" type="remote">
  <instance_attributes id="{self._remote_node}-instance_attributes">
    <nvpair id="{self._remote_node}-instance_attributes-server" name="server" value="{node}"/>
"""

        if self._remote_use_reconnect_interval:
            # Set reconnect interval on resource
            rsc_xml += f"""
    <nvpair id="{self._remote_node}-instance_attributes-reconnect_interval" name="reconnect_interval" value="60s"/>
"""

        rsc_xml += f"""
  </instance_attributes>
  <operations>
    <op id="{self._remote_node}-start"       name="start"   interval="0"   timeout="120s"/>
    <op id="{self._remote_node}-monitor-20s" name="monitor" interval="20s" timeout="45s"/>
  </operations>
</primitive>
"""

        self._add_rsc(node, rsc_xml)
        if not self.failed:
            self._remote_node_added = True

    def _disable_services(self, node):
        """Disable the corosync and pacemaker services on the given node."""
        self._corosync_enabled = self._env.service_is_enabled(node, "corosync")
        if self._corosync_enabled:
            self._env.disable_service(node, "corosync")

        self._pacemaker_enabled = self._env.service_is_enabled(node, "pacemaker")
        if self._pacemaker_enabled:
            self._env.disable_service(node, "pacemaker")

    def _enable_services(self, node):
        """Enable the corosync and pacemaker services on the given node."""
        if self._corosync_enabled:
            self._env.enable_service(node, "corosync")

        if self._pacemaker_enabled:
            self._env.enable_service(node, "pacemaker")

    def _stop_pcmk_remote(self, node):
        """Stop the Pacemaker Remote service on the given node."""
        for _ in range(10):
            (rc, _) = self._rsh(node, "service pacemaker_remote stop")
            if rc != 0:
                time.sleep(6)
            else:
                break

    def _start_pcmk_remote(self, node):
        """Start the Pacemaker Remote service on the given node."""
        for _ in range(10):
            (rc, _) = self._rsh(node, "service pacemaker_remote start")
            if rc != 0:
                time.sleep(6)
            else:
                self._pcmk_started = True
                break

    def _freeze_pcmk_remote(self, node):
        """Simulate a Pacemaker Remote daemon failure."""
        self._rsh(node, "killall -STOP pacemaker-remoted")

    def _resume_pcmk_remote(self, node):
        """Simulate the Pacemaker Remote daemon recovering."""
        self._rsh(node, "killall -CONT pacemaker-remoted")

    def _start_metal(self, node):
        """
        Set up a Pacemaker Remote configuration.

        Remove any existing connection resources or nodes.  Start the
        pacemaker_remote service.  Create a connection resource.
        """
        # Cluster nodes are reused as remote nodes in remote tests. If cluster
        # services were enabled at boot, in case the remote node got fenced, the
        # cluster node would join instead of the expected remote one. Meanwhile
        # pacemaker_remote would not be able to start. Depending on the chances,
        # the situations might not be able to be orchestrated gracefully any more.
        #
        # Temporarily disable any enabled cluster serivces.
        self._disable_services(node)

        # make sure the resource doesn't already exist for some reason
        self._rsh(node, f"crm_resource -D -r {self._remote_rsc} -t primitive")
        self._rsh(node, f"crm_resource -D -r {self._remote_node} -t primitive")

        if not self._stop(node):
            self.fail(f"Failed to shutdown cluster node {node}")
            return

        self._start_pcmk_remote(node)

        if not self._pcmk_started:
            self.fail(f"Failed to start pacemaker_remote on node {node}")
            return

        # Convert node to baremetal now that it has shutdown the cluster stack
        pats = []
        watch = self.create_watch(pats, 120)
        watch.set_watch()

        pats.extend([
            self.templates["Pat:RscOpOK"] % ("start", self._remote_node),
            self.templates["Pat:DC_IDLE"]
        ])

        self._add_connection_rsc(node)

        with Timer(self._logger, self.name, "remoteMetalInit"):
            watch.look_for_all()

        if watch.unmatched:
            self.fail(f"Unmatched patterns: {watch.unmatched}")

    def migrate_connection(self, node):
        """Move the remote connection resource to any other available node."""
        if self.failed:
            return

        pats = [
            self.templates["Pat:RscOpOK"] % ("migrate_to", self._remote_node),
            self.templates["Pat:RscOpOK"] % ("migrate_from", self._remote_node),
            self.templates["Pat:DC_IDLE"]
        ]

        watch = self.create_watch(pats, 120)
        watch.set_watch()

        (rc, _) = self._rsh(node, f"crm_resource -M -r {self._remote_node}", verbose=1)
        if rc != 0:
            self.fail("failed to move remote node connection resource")
            return

        with Timer(self._logger, self.name, "remoteMetalMigrate"):
            watch.look_for_all()

        if watch.unmatched:
            self.fail(f"Unmatched patterns: {watch.unmatched}")

    def fail_rsc(self, node):
        """
        Cause the dummy resource running on a Pacemaker Remote node to fail.

        Verify that the failure is logged correctly.
        """
        if self.failed:
            return

        watchpats = [
            self.templates["Pat:RscRemoteOpOK"] % ("stop", self._remote_rsc, self._remote_node),
            self.templates["Pat:RscRemoteOpOK"] % ("start", self._remote_rsc, self._remote_node),
            self.templates["Pat:DC_IDLE"]
        ]

        watch = self.create_watch(watchpats, 120)
        watch.set_watch()

        self.debug("causing dummy rsc to fail.")

        self._rsh(node, "rm -f /var/run/resource-agents/Dummy*")

        with Timer(self._logger, self.name, "remoteRscFail"):
            watch.look_for_all()

        if watch.unmatched:
            self.fail(f"Unmatched patterns during rsc fail: {watch.unmatched}")

    def fail_connection(self, node):
        """
        Cause the remote connection resource to fail.

        Verify that the node is fenced and the connection resource is restarted
        on another node.
        """
        if self.failed:
            return

        watchpats = [
            self.templates["Pat:Fencing_ok"] % self._remote_node,
            self.templates["Pat:NodeFenced"] % self._remote_node
        ]

        watch = self.create_watch(watchpats, 120)
        watch.set_watch()

        # freeze the pcmk remote daemon. this will result in fencing
        self.debug("Force stopped active remote node")
        self._freeze_pcmk_remote(node)

        self.debug("Waiting for remote node to be fenced.")

        with Timer(self._logger, self.name, "remoteMetalFence"):
            watch.look_for_all()

        if watch.unmatched:
            self.fail(f"Unmatched patterns: {watch.unmatched}")
            return

        self.debug("Waiting for the remote node to come back up")
        self._cm.ns.wait_for_node(node, 120)

        pats = []

        watch = self.create_watch(pats, 240)
        watch.set_watch()

        pats.append(self.templates["Pat:RscOpOK"] % ("start", self._remote_node))

        if self._remote_rsc_added:
            pats.append(self.templates["Pat:RscRemoteOpOK"] % ("start", self._remote_rsc, self._remote_node))

        # start the remote node again watch it integrate back into cluster.
        self._start_pcmk_remote(node)
        if not self._pcmk_started:
            self.fail(f"Failed to start pacemaker_remote on node {node}")
            return

        self.debug("Waiting for remote node to rejoin cluster after being fenced.")

        with Timer(self._logger, self.name, "remoteMetalRestart"):
            watch.look_for_all()

        if watch.unmatched:
            self.fail(f"Unmatched patterns: {watch.unmatched}")

    def _add_dummy_rsc(self, node):
        """Add a dummy resource that runs on the Pacemaker Remote node."""
        if self.failed:
            return

        # verify we can put a resource on the remote node
        pats = []
        watch = self.create_watch(pats, 120)
        watch.set_watch()

        pats.extend([
            self.templates["Pat:RscRemoteOpOK"] % ("start", self._remote_rsc, self._remote_node),
            self.templates["Pat:DC_IDLE"]
        ])

        # Add a resource that must live on remote-node
        self._add_primitive_rsc(node)

        # force that rsc to prefer the remote node.
        (rc, _) = self._cm.rsh(node, f"crm_resource -M -r {self._remote_rsc} -N {self._remote_node} -f", verbose=1)
        if rc != 0:
            self.fail("Failed to place remote resource on remote node.")
            return

        with Timer(self._logger, self.name, "remoteMetalRsc"):
            watch.look_for_all()

        if watch.unmatched:
            self.fail(f"Unmatched patterns: {watch.unmatched}")

    def test_attributes(self, node):
        """Verify that attributes can be set on the Pacemaker Remote node."""
        if self.failed:
            return

        # This verifies permanent attributes can be set on a remote-node. It also
        # verifies the remote-node can edit its own cib node section remotely.
        (rc, line) = self._cm.rsh(node, f"crm_attribute -l forever -n testattr -v testval -N {self._remote_node}", verbose=1)
        if rc != 0:
            self.fail(f"Failed to set remote-node attribute. rc:{rc} output:{line}")
            return

        (rc, _) = self._cm.rsh(node, f"crm_attribute -l forever -n testattr -q -N {self._remote_node}", verbose=1)
        if rc != 0:
            self.fail("Failed to get remote-node attribute")
            return

        (rc, _) = self._cm.rsh(node, f"crm_attribute -l forever -n testattr -D -N {self._remote_node}", verbose=1)
        if rc != 0:
            self.fail("Failed to delete remote-node attribute")

    def cleanup_metal(self, node):
        """
        Clean up the Pacemaker Remote node configuration previously created by _setup_metal.

        Stop and remove dummy resources and connection resources.  Stop the
        pacemaker_remote service.  Remove the remote node itself.
        """
        self._enable_services(node)

        if not self._pcmk_started:
            return

        pats = []

        watch = self.create_watch(pats, 120)
        watch.set_watch()

        if self._remote_rsc_added:
            pats.append(self.templates["Pat:RscOpOK"] % ("stop", self._remote_rsc))

        if self._remote_node_added:
            pats.append(self.templates["Pat:RscOpOK"] % ("stop", self._remote_node))

        with Timer(self._logger, self.name, "remoteMetalCleanup"):
            self._resume_pcmk_remote(node)

            if self._remote_rsc_added:
                # Remove dummy resource added for remote node tests
                self.debug("Cleaning up dummy rsc put on remote node")
                self._rsh(self._get_other_node(node), f"crm_resource -U -r {self._remote_rsc}")
                self._del_rsc(node, self._remote_rsc)

            if self._remote_node_added:
                # Remove remote node's connection resource
                self.debug("Cleaning up remote node connection resource")
                self._rsh(self._get_other_node(node), f"crm_resource -U -r {self._remote_node}")
                self._del_rsc(node, self._remote_node)

            watch.look_for_all()

        if watch.unmatched:
            self.fail(f"Unmatched patterns: {watch.unmatched}")

        self._stop_pcmk_remote(node)

        self.debug("Waiting for the cluster to recover")
        self._cm.cluster_stable()

        if self._remote_node_added:
            # Remove remote node itself
            self.debug("Cleaning up node entry for remote node")
            self._rsh(self._get_other_node(node), f"crm_node --force --remove {self._remote_node}")

    def _setup_env(self, node):
        """
        Set up the environment to allow Pacemaker Remote to function.

        This involves generating a key and copying it to all nodes in the cluster.
        """
        self._remote_node = f"remote-{node}"

        # we are assuming if all nodes have a key, that it is
        # the right key... If any node doesn't have a remote
        # key, we regenerate it everywhere.
        if self._rsh.exists_on_all("/etc/pacemaker/authkey", self._env["nodes"]):
            return

        # create key locally
        (handle, keyfile) = tempfile.mkstemp(".cts")
        os.close(handle)
        subprocess.check_call(["dd", "if=/dev/urandom", f"of={keyfile}", "bs=4096", "count=1"],
                              stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

        # sync key throughout the cluster
        for n in self._env["nodes"]:
            self._rsh(n, "mkdir -p --mode=0750 /etc/pacemaker")
            self._rsh.copy(keyfile, f"root@{n}:/etc/pacemaker/authkey")
            self._rsh(n, "chgrp haclient /etc/pacemaker /etc/pacemaker/authkey")
            self._rsh(n, "chmod 0640 /etc/pacemaker/authkey")

        os.unlink(keyfile)

    def is_applicable(self):
        """Return True if this test is applicable in the current test configuration."""
        if not CTSTest.is_applicable(self):
            return False

        for node in self._env["nodes"]:
            (rc, _) = self._rsh(node, "which pacemaker-remoted >/dev/null 2>&1")
            if rc != 0:
                return False

        return True

    def start_new_test(self, node):
        """Prepare a remote test for running by setting up its environment and resources."""
        self.incr("calls")
        self.reset()

        ret = self._startall(None)
        if not ret:
            return self.failure("setup failed: could not start all nodes")

        self._setup_env(node)
        self._start_metal(node)
        self._add_dummy_rsc(node)
        return True

    def __call__(self, node):
        """Perform this test."""
        raise NotImplementedError

    @property
    def errors_to_ignore(self):
        """Return list of errors which should be ignored."""
        return [
            r"""is running on remote.*which isn't allowed""",
            r"""Connection terminated""",
            r"""Could not send remote"""
        ]



"""Randomly start and stop nodes to bring the cluster close to the quorum point."""

__all__ = ["NearQuorumPointTest"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.ctstest import CTSTest

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._rsh is callable.
# pylint: disable=not-callable
# pylint doesn't understand that self._env is subscriptable.
# pylint: disable=unsubscriptable-object


class NearQuorumPointTest(CTSTest):
    """Randomly start and stop nodes to bring the cluster close to the quorum point."""

    def __init__(self, cm):
        """
        Create a new NearQuorumPointTest instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.name = "NearQuorumPoint"

    def __call__(self, dummy):
        """Perform this test."""
        self.incr("calls")
        startset = []
        stopset = []

        stonith = self._cm.prepare_fencing_watcher()
        # decide what to do with each node
        for node in self._env["nodes"]:
            action = self._env.random_gen.choice(["start", "stop"])

            if action == "start":
                startset.append(node)
            elif action == "stop":
                stopset.append(node)

        self.debug(f"start nodes:{startset!r}")
        self.debug(f"stop nodes:{stopset!r}")

        # add search patterns
        watchpats = []
        for node in stopset:
            if self._cm.expected_status[node] == "up":
                watchpats.append(self.templates["Pat:We_stopped"] % node)

        for node in startset:
            if self._cm.expected_status[node] == "down":
                watchpats.append(self.templates["Pat:Local_started"] % node)
            else:
                for stopping in stopset:
                    if self._cm.expected_status[stopping] == "up":
                        watchpats.append(self.templates["Pat:They_stopped"] % (node, stopping))

        if not watchpats:
            return self.skipped()

        if startset:
            watchpats.append(self.templates["Pat:DC_IDLE"])

        watch = self.create_watch(watchpats, self._env["DeadTime"] + 10)

        watch.set_watch()

        # begin actions
        for node in stopset:
            if self._cm.expected_status[node] == "up":
                self._cm.stop_cm_async(node)

        for node in startset:
            if self._cm.expected_status[node] == "down":
                self._cm.start_cm_async(node)

        # get the result
        if watch.look_for_all():
            self._cm.cluster_stable()
            self._cm.fencing_cleanup("NearQuorumPoint", stonith)
            return self.success()

        self._logger.log(f"Warn: Patterns not found: {watch.unmatched!r}")

        # get the "bad" nodes
        upnodes = []
        for node in stopset:
            if self._cm.stat_cm(node):
                upnodes.append(node)

        downnodes = []
        for node in startset:
            if not self._cm.stat_cm(node):
                downnodes.append(node)

        self._cm.fencing_cleanup("NearQuorumPoint", stonith)
        if not upnodes and not downnodes:
            self._cm.cluster_stable()

            # Make sure they're completely down with no residule
            for node in stopset:
                self._rsh(node, self.templates["StopCmd"])

            return self.success()

        if upnodes:
            self._logger.log(f"Warn: Unstoppable nodes: {upnodes!r}")

        if downnodes:
            self._logger.log(f"Warn: Unstartable nodes: {downnodes!r}")

        return self.failure()



"""A module providing functions for manipulating corosync."""

__all__ = ["Corosync", "localname"]
__copyright__ = "Copyright 2009-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+)"

import os
import shutil
import subprocess
import tempfile
import time

from pacemaker.buildoptions import BuildOptions
from pacemaker._cts.process import killall, stdout_from_command


AUTOGEN_COROSYNC_TEMPLATE = """
totem {
    version: 2
    cluster_name: %s
    crypto_cipher: none
    crypto_hash: none
    transport: udp
}

nodelist {
    node {
        nodeid: 1
        name: %s
        ring0_addr: 127.0.0.1
    }
}

logging {
    debug: off
    to_syslog: no
    to_stderr: no
    to_logfile: yes
    logfile: %s
}
"""


def corosync_cfg_exists():
    """Return whether the corosync config file exists."""
    return os.path.exists(BuildOptions.COROSYNC_CONFIG_FILE)


def corosync_log_file(cfgfile):
    """Return the path to the corosync log file, or None."""
    with open(cfgfile, "r", encoding="utf-8") as f:
        for line in f.readlines():
            # "to_logfile:" could also be in the config file, so check for a
            # slash to make sure it's a path we're looking at.
            if "logfile: /" in line:
                return line.split()[-1]

    return None


def generate_corosync_cfg(logdir, cluster_name, node_name):
    """
    Generate a corosync config file.

    If there's a corosync config file already installed on the system, move
    it to a temporary location and return that temporary name.  Otherwise,
    return None.
    """
    retval = None

    if corosync_cfg_exists():
        # pylint: disable=consider-using-with
        config_dir = os.path.dirname(BuildOptions.COROSYNC_CONFIG_FILE)
        f = tempfile.NamedTemporaryFile(dir=config_dir, prefix="corosync.conf-")
        f.close()
        shutil.move(BuildOptions.COROSYNC_CONFIG_FILE, f.name)

        retval = f.name

    logfile = os.path.join(logdir, "corosync.log")

    with open(BuildOptions.COROSYNC_CONFIG_FILE, "w", encoding="utf-8") as corosync_cfg:
        corosync_cfg.write(AUTOGEN_COROSYNC_TEMPLATE % (cluster_name, node_name, logfile))

    return retval


def localname():
    """Return the uname of the local host."""
    our_uname = stdout_from_command(["uname", "-n"])
    if our_uname:
        our_uname = our_uname[0]
    else:
        our_uname = "localhost"

    return our_uname


class Corosync:
    """A class for managing corosync processes and config files."""

    def __init__(self, verbose, logdir, cluster_name):
        """
        Create a new Corosync instance.

        Arguments:
        verbose      -- Whether to print the corosync log file
        logdir       -- The base directory under which to store log files
        cluster_name -- The name of the cluster
        """
        self.verbose = verbose
        self.logdir = logdir
        self.cluster_name = cluster_name

        self._existing_cfg_file = None

    def _ready(self, logfile, timeout=10):
        """Return whether corosync is ready."""
        i = 0

        while i < timeout:
            with open(logfile, "r", encoding="utf-8") as corosync_log:
                for line in corosync_log.readlines():
                    if line.endswith("ready to provide service.\n"):
                        # Even once the line is in the log file, we may still need to wait just
                        # a little bit longer before corosync is really ready to go.
                        time.sleep(1)
                        return

            time.sleep(1)
            i += 1

        raise TimeoutError

    def start(self, kill_first=False, timeout=10):
        """
        Start the corosync process.

        Arguments:
        kill_first -- Whether to kill any pre-existing corosync processes before
                      starting a new one
        timeout    -- If corosync does not start within this many seconds, raise
                      TimeoutError
        """
        if kill_first:
            killall(["corosync"])

        self._existing_cfg_file = generate_corosync_cfg(self.logdir,
                                                        self.cluster_name, localname())
        logfile = corosync_log_file(BuildOptions.COROSYNC_CONFIG_FILE)

        if self.verbose:
            print("Starting corosync")

        with subprocess.Popen("corosync", stdout=subprocess.PIPE, stderr=subprocess.PIPE) as test:
            test.wait()

        # Wait for corosync to be ready before returning
        self._ready(logfile, timeout=timeout)

    def stop(self):
        """Stop the corosync process."""
        killall(["corosync"])

        if self.verbose:
            print("Corosync output")

            logfile = corosync_log_file(BuildOptions.COROSYNC_CONFIG_FILE)
            with open(logfile, "r", encoding="utf-8") as corosync_log:
                for line in corosync_log.readlines():
                    print(line.strip())

        os.remove(BuildOptions.COROSYNC_CONFIG_FILE)

        # If there was a previous corosync config file, move it back into place
        if self._existing_cfg_file:
            shutil.move(self._existing_cfg_file, BuildOptions.COROSYNC_CONFIG_FILE)



"""Auditing classes for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = ["AuditConstraint", "AuditResource", "ClusterAudit", "audit_list"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import re
import time
import uuid

from pacemaker.buildoptions import BuildOptions
from pacemaker._cts.input import should_continue
from pacemaker._cts.watcher import LogKind, LogWatcher


class ClusterAudit:
    """
    The base class for various kinds of auditors.

    Specific audit implementations should be built on top of this one.  Audits
    can do all kinds of checks on the system.  The basic interface for callers
    is the `__call__` method, which returns True if the audit passes and False
    if it fails.
    """

    def __init__(self, cm):
        """
        Create a new ClusterAudit instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        # pylint: disable=invalid-name
        self._cm = cm
        self.name = None

    def __call__(self):
        """Perform the audit action."""
        raise NotImplementedError

    def is_applicable(self):
        """
        Return True if this audit is applicable in the current test configuration.

        This method must be implemented by all subclasses.
        """
        raise NotImplementedError

    def log(self, args):
        """Log a message."""
        self._cm.log(f"audit: {args}")

    def debug(self, args):
        """Log a debug message."""
        self._cm.debug(f"audit: {args}")


class LogAudit(ClusterAudit):
    """
    Audit each cluster node to verify that some logging system is usable.

    This is done by logging a unique test message and then verifying that we
    can read back that test message using logging tools.
    """

    def __init__(self, cm):
        """
        Create a new LogAudit instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        ClusterAudit.__init__(self, cm)
        self.name = "LogAudit"

    def _restart_cluster_logging(self, nodes=None):
        """Restart logging on the given nodes, or all if none are given."""
        if not nodes:
            nodes = self._cm.env["nodes"]

        self._cm.debug(f"Restarting logging on: {nodes!r}")

        for node in nodes:
            if self._cm.env["have_systemd"]:
                (rc, _) = self._cm.rsh(node, "systemctl stop systemd-journald.socket")
                if rc != 0:
                    self._cm.log(f"ERROR: Cannot stop 'systemd-journald' on {node}")

                (rc, _) = self._cm.rsh(node, "systemctl start systemd-journald.service")
                if rc != 0:
                    self._cm.log(f"ERROR: Cannot start 'systemd-journald' on {node}")

            if "syslogd" in self._cm.env:
                (rc, _) = self._cm.rsh(node, f"service {self._cm.env['syslogd']} restart")
                if rc != 0:
                    self._cm.log(f"""ERROR: Cannot restart '{self._cm.env["syslogd"]}' on {node}""")

    def _create_watcher(self, patterns, kind):
        """Create a new LogWatcher instance for the given patterns."""
        watch = LogWatcher(self._cm.env["LogFileName"], patterns,
                           self._cm.env["nodes"], kind, "LogAudit", 5,
                           silent=True)
        watch.set_watch()
        return watch

    def _test_logging(self):
        """Perform the log audit."""
        patterns = []
        prefix = "Test message from"
        suffix = str(uuid.uuid4())
        watch = {}

        for node in self._cm.env["nodes"]:
            # Look for the node name in two places to make sure
            # that syslog is logging with the correct hostname
            m = re.search("^([^.]+).*", node)
            if m:
                simple = m.group(1)
            else:
                simple = node

            patterns.append(f"{simple}.*{prefix} {node} {suffix}")

        watch_pref = self._cm.env["log_kind"]
        if watch_pref is None:
            kinds = [LogKind.LOCAL_FILE]
            if self._cm.env["have_systemd"]:
                kinds.append(LogKind.JOURNAL)
            kinds.append(LogKind.REMOTE_FILE)

            for k in kinds:
                watch[k] = self._create_watcher(patterns, k)
            self._cm.log(f"Logging test message with identifier {suffix}")
        else:
            watch[watch_pref] = self._create_watcher(patterns, watch_pref)

        for node in self._cm.env["nodes"]:
            cmd = f"logger -p {self._cm.env['SyslogFacility']}.info {prefix} {node} {suffix}"

            (rc, _) = self._cm.rsh(node, cmd, synchronous=False, verbose=0)
            if rc != 0:
                self._cm.log(f"ERROR: Cannot execute remote command [{cmd}] on {node}")

        for k in list(watch.keys()):
            w = watch[k]
            if watch_pref is None:
                self._cm.log(f"Checking for test message in {k} logs")
            w.look_for_all(silent=True)
            if w.unmatched:
                for regex in w.unmatched:
                    self._cm.log(f"Test message [{regex}] not found in {w.kind} logs")
            else:
                if watch_pref is None:
                    self._cm.log(f"Found test message in {k} logs")
                    self._cm.env["log_kind"] = k
                return 1

        return False

    def __call__(self):
        """Perform the audit action."""
        max_attempts = 3
        attempt = 0
        passed = True

        self._cm.ns.wait_for_all_nodes(self._cm.env["nodes"])
        while attempt <= max_attempts and not self._test_logging():
            attempt += 1
            self._restart_cluster_logging()
            time.sleep(60 * attempt)

        if attempt > max_attempts:
            self._cm.log("ERROR: Cluster logging unrecoverable.")
            passed = False

        return passed

    def is_applicable(self):
        """Return True if this audit is applicable in the current test configuration."""
        if self._cm.env["LogAuditDisabled"]:
            return False

        return True


class DiskAudit(ClusterAudit):
    """
    Audit disk usage on cluster nodes.

    Verify that there is enough free space left on whichever mounted file
    system holds the logs.

    Warn on:  less than 100 MB or 10% of free space
    Error on: less than 10 MB or 5% of free space
    """

    def __init__(self, cm):
        """
        Create a new DiskAudit instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        ClusterAudit.__init__(self, cm)
        self.name = "DiskspaceAudit"

    def __call__(self):
        """Perform the audit action."""
        passed = True

        # @TODO Use directory of PCMK_logfile if set on host
        dfcmd = "df -BM %s | tail -1 | awk '{print $(NF-1)\" \"$(NF-2)}' | tr -d 'M%%'" % BuildOptions.LOG_DIR

        self._cm.ns.wait_for_all_nodes(self._cm.env["nodes"])
        for node in self._cm.env["nodes"]:
            (_, dfout) = self._cm.rsh(node, dfcmd, verbose=1)
            if not dfout:
                self._cm.log(f"ERROR: Cannot execute remote df command [{dfcmd}] on {node}")
                continue

            dfout = dfout[0].strip()

            try:
                (used, remain) = dfout.split()
                used_percent = int(used)
                remaining_mb = int(remain)
            except (ValueError, TypeError):
                self._cm.log(f"Warning: df output '{dfout}' from {node} was invalid [{used}, {remain}]")
            else:
                if remaining_mb < 10 or used_percent > 95:
                    self._cm.log(f"CRIT: Out of log disk space on {node} ({used_percent}% / {remaining_mb}MB)")
                    passed = False

                    if not should_continue(self._cm.env):
                        raise ValueError(f"Disk full on {node}")

                elif remaining_mb < 100 or used_percent > 90:
                    self._cm.log(f"WARN: Low on log disk space ({remaining_mb}MB) on {node}")

        return passed

    def is_applicable(self):
        """Return True if this audit is applicable in the current test configuration."""
        return True


class FileAudit(ClusterAudit):
    """
    Audit the filesystem looking for various failure conditions.

    Check for:
    * The presence of core dumps from corosync or Pacemaker daemons
    * Stale IPC files
    """

    def __init__(self, cm):
        """
        Create a new FileAudit instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        ClusterAudit.__init__(self, cm)
        self.known = []
        self.name = "FileAudit"

    def _output_has_core(self, output, node):
        """Check output for any lines that would indicate the presence of a core dump."""
        found = False

        for line in output:
            line = line.strip()

            if line in self.known:
                continue

            found = True
            self.known.append(line)
            self._cm.log(f"Warning: core file on {node}: {line}")

        return found

    def _find_core_with_coredumpctl(self, node):
        """Use coredumpctl to find core dumps on the given node."""
        (_, lsout) = self._cm.rsh(node, "coredumpctl --no-legend --no-pager")
        return self._output_has_core(lsout, node)

    def _find_core_on_fs(self, node, paths):
        """Check for core dumps on the given node, under any of the given paths."""
        (_, lsout) = self._cm.rsh(node, f"ls -al {' '.join(paths)} | grep core.[0-9]",
                                  verbose=1)
        return self._output_has_core(lsout, node)

    def __call__(self):
        """Perform the audit action."""
        passed = True

        self._cm.ns.wait_for_all_nodes(self._cm.env["nodes"])

        for node in self._cm.env["nodes"]:
            found = False

            # If systemd is present, first see if coredumpctl logged any core dumps.
            if self._cm.env["have_systemd"]:
                found = self._find_core_with_coredumpctl(node)
                if found:
                    passed = False

            # If we didn't find any core dumps, it's for one of three reasons:
            # (1) Nothing crashed
            # (2) systemd is not present
            # (3) systemd is present but coredumpctl is not enabled
            #
            # To handle the last two cases, check the other filesystem locations.
            if not found:
                found = self._find_core_on_fs(node, ["/var/lib/pacemaker/cores/*",
                                                     "/var/lib/corosync"])
                if found:
                    passed = False

            if self._cm.expected_status.get(node) == "down":
                clean = False
                (_, lsout) = self._cm.rsh(node, "ls -al /dev/shm | grep qb-", verbose=1)

                for line in lsout:
                    passed = False
                    clean = True
                    self._cm.log(f"Warning: Stale IPC file on {node}: {line}")

                if clean:
                    (_, lsout) = self._cm.rsh(node, "ps axf | grep -e pacemaker -e corosync", verbose=1)

                    for line in lsout:
                        self._cm.debug(f"ps[{node}]: {line}")

                    self._cm.rsh(node, "rm -rf /dev/shm/qb-*")

            else:
                self._cm.debug(f"Skipping {node}")

        return passed

    def is_applicable(self):
        """Return True if this audit is applicable in the current test configuration."""
        return True


class AuditResource:
    """A base class for storing information about a cluster resource."""

    def __init__(self, cm, line):
        """
        Create a new AuditResource instance.

        Arguments:
        cm   -- A ClusterManager instance
        line -- One line of output from `crm_resource` describing a single
                resource
        """
        # pylint: disable=invalid-name
        fields = line.split()
        self._cm = cm
        self.line = line
        self.type = fields[1]
        self.id = fields[2]
        self.clone_id = fields[3]
        self.parent = fields[4]
        self.rprovider = fields[5]
        self.rclass = fields[6]
        self.rtype = fields[7]
        self.host = fields[8]
        self.needs_quorum = fields[9]
        self.flags = int(fields[10])
        self.flags_s = fields[11]

        if self.parent == "NA":
            self.parent = None

    @property
    def unique(self):
        """Return True if this resource is unique."""
        return self.flags & 0x20

    @property
    def orphan(self):
        """Return True if this resource is an orphan."""
        return self.flags & 0x01

    @property
    def managed(self):
        """Return True if this resource is managed by the cluster."""
        return self.flags & 0x02


class AuditConstraint:
    """A base class for storing information about a cluster constraint."""

    def __init__(self, cm, line):
        """
        Create a new AuditConstraint instance.

        Arguments:
        cm   -- A ClusterManager instance
        line -- One line of output from `crm_resource` describing a single
                constraint
        """
        # pylint: disable=invalid-name
        fields = line.split()
        self._cm = cm
        self.line = line
        self.type = fields[1]
        self.id = fields[2]
        self.rsc = fields[3]
        self.target = fields[4]
        self.score = fields[5]
        self.rsc_role = fields[6]
        self.target_role = fields[7]

        if self.rsc_role == "NA":
            self.rsc_role = None

        if self.target_role == "NA":
            self.target_role = None


class PrimitiveAudit(ClusterAudit):
    """
    Audit primitive resources to verify a variety of conditions.

    Check that:
    * Resources are active and managed only when expected
    * Resources are active on the expected cluster node
    * Resources are not orphaned
    """

    def __init__(self, cm):
        """
        Create a new PrimitiveAudit instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        ClusterAudit.__init__(self, cm)
        self.name = "PrimitiveAudit"

        self._active_nodes = []
        self._constraints = []
        self._inactive_nodes = []
        self._resources = []
        self._target = None

    def _audit_resource(self, resource, quorum):
        """Perform the audit of a single resource."""
        rc = True
        active = self._cm.resource_location(resource.id)

        if len(active) == 1:
            if quorum:
                self.debug(f"Resource {resource.id} active on {active!r}")

            elif resource.needs_quorum == 1:
                self._cm.log(f"Resource {resource.id} active without quorum: {active!r}")
                rc = False

        elif not resource.managed:
            self._cm.log(f"Resource {resource.id} not managed. Active on {active!r}")

        elif not resource.unique:
            # TODO: Figure out a clever way to actually audit these resource types
            if len(active) > 1:
                self.debug(f"Non-unique resource {resource.id} is active on: {active!r}")
            else:
                self.debug(f"Non-unique resource {resource.id} is not active")

        elif len(active) > 1:
            self._cm.log(f"Resource {resource.id} is active multiple times: {active!r}")
            rc = False

        elif resource.orphan:
            self.debug(f"Resource {resource.id} is an inactive orphan")

        elif not self._inactive_nodes:
            self._cm.log(f"WARN: Resource {resource.id} not served anywhere")
            rc = False

        elif self._cm.env["warn-inactive"]:
            if quorum or not resource.needs_quorum:
                self._cm.log(f"WARN: Resource {resource.id} not served anywhere "
                             f"(Inactive nodes: {self._inactive_nodes!r})")
            else:
                self.debug(f"Resource {resource.id} not served anywhere "
                           f"(Inactive nodes: {self._inactive_nodes!r})")

        elif quorum or not resource.needs_quorum:
            self.debug(f"Resource {resource.id} not served anywhere "
                       f"(Inactive nodes: {self._inactive_nodes!r})")

        return rc

    def _setup(self):
        """
        Verify cluster nodes are active.

        Collect resource and colocation information used for performing the audit.
        """
        for node in self._cm.env["nodes"]:
            if self._cm.expected_status[node] == "up":
                self._active_nodes.append(node)
            else:
                self._inactive_nodes.append(node)

        for node in self._cm.env["nodes"]:
            if self._target is None and self._cm.expected_status[node] == "up":
                self._target = node

        if not self._target:
            # TODO: In Pacemaker 1.0 clusters we'll be able to run crm_resource
            # with CIB_file=/path/to/cib.xml even when the cluster isn't running
            self.debug(f"No nodes active - skipping {self.name}")
            return False

        (_, lines) = self._cm.rsh(self._target, "crm_resource --list-cts",
                                  verbose=1)

        for line in lines:
            if re.search("^Resource", line):
                self._resources.append(AuditResource(self._cm, line))
            elif re.search("^Constraint", line):
                self._constraints.append(AuditConstraint(self._cm, line))
            else:
                self._cm.log(f"Unknown entry: {line}")

        return True

    def __call__(self):
        """Perform the audit action."""
        passed = True

        if not self._setup():
            return passed

        quorum = self._cm.has_quorum(None)
        for resource in self._resources:
            if resource.type == "primitive" and not self._audit_resource(resource, quorum):
                passed = False

        return passed

    def is_applicable(self):
        """Return True if this audit is applicable in the current test configuration."""
        # @TODO Due to long-ago refactoring, this name test would never match,
        # so this audit (and those derived from it) would never run.
        # Uncommenting the next lines fixes the name test, but that then
        # exposes pre-existing bugs that need to be fixed.
        # if self._cm["Name"] == "crm-corosync":
        #     return True
        return False


class GroupAudit(PrimitiveAudit):
    """
    Audit group resources.

    Check that:
    * Each of its child primitive resources is active on the expected cluster node
    """

    def __init__(self, cm):
        """
        Create a new GroupAudit instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        PrimitiveAudit.__init__(self, cm)
        self.name = "GroupAudit"

    def __call__(self):
        passed = True

        if not self._setup():
            return passed

        for group in self._resources:
            if group.type != "group":
                continue

            first_match = True
            group_location = None

            for child in self._resources:
                if child.parent != group.id:
                    continue

                nodes = self._cm.resource_location(child.id)

                if first_match and len(nodes) > 0:
                    group_location = nodes[0]

                first_match = False

                if len(nodes) > 1:
                    passed = False
                    self._cm.log(f"Child {child.id} of {group.id} is active more than once: {nodes!r}")

                elif not nodes:
                    # Groups are allowed to be partially active
                    # However we do need to make sure later children aren't running
                    group_location = None
                    self.debug(f"Child {child.id} of {group.id} is stopped")

                elif nodes[0] != group_location:
                    passed = False
                    self._cm.log(f"Child {child.id} of {group.id} is active on the wrong "
                                 f"node ({nodes[0]}) expected {group_location}")
                else:
                    self.debug(f"Child {child.id} of {group.id} is active on {nodes[0]}")

        return passed


class CloneAudit(PrimitiveAudit):
    """
    Audit clone resources.

    NOTE: Currently, this class does not perform any actual audit functions.
    """

    def __init__(self, cm):
        """
        Create a new CloneAudit instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        PrimitiveAudit.__init__(self, cm)
        self.name = "CloneAudit"

    def __call__(self):
        passed = True

        if not self._setup():
            return passed

        for clone in self._resources:
            if clone.type != "clone":
                continue

            for child in self._resources:
                if child.parent == clone.id and child.type == "primitive":
                    self.debug(f"Checking child {child.id} of {clone.id}...")
                    # Check max and node_max
                    # Obtain with:
                    #    crm_resource -g clone_max --meta -r child.id
                    #    crm_resource -g clone_node_max --meta -r child.id

        return passed


class ColocationAudit(PrimitiveAudit):
    """
    Audit cluster resources.

    Check that:

    * Resources are colocated with the expected resource
    """

    def __init__(self, cm):
        """
        Create a new ColocationAudit instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        PrimitiveAudit.__init__(self, cm)
        self.name = "ColocationAudit"

    def _crm_location(self, resource):
        """Return a list of cluster nodes where a given resource is running."""
        (rc, lines) = self._cm.rsh(self._target,
                                   f"crm_resource --locate -r {resource} -Q",
                                   verbose=1)
        hosts = []

        if rc == 0:
            for line in lines:
                fields = line.split()
                hosts.append(fields[0])

        return hosts

    def __call__(self):
        passed = True

        if not self._setup():
            return passed

        for coloc in self._constraints:
            if coloc.type != "rsc_colocation":
                continue

            source = self._crm_location(coloc.rsc)
            target = self._crm_location(coloc.target)

            if not source:
                self.debug(f"Colocation audit ({coloc.id}): {coloc.rsc} not running")
            else:
                for node in source:
                    if node not in target:
                        passed = False
                        self._cm.log(f"Colocation audit ({coloc.id}): {coloc.rsc} running "
                                     f"on {node} (not in {target!r})")
                    else:
                        self.debug(f"Colocation audit ({coloc.id}): {coloc.rsc} running "
                                   f"on {node} (in {target!r})")

        return passed


class ControllerStateAudit(ClusterAudit):
    """Verify active and inactive resources."""

    def __init__(self, cm):
        """
        Create a new ControllerStateAudit instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        ClusterAudit.__init__(self, cm)
        self.name = "ControllerStateAudit"

    def __call__(self):
        passed = True
        up_are_down = 0
        down_are_up = 0
        unstable_list = []

        for node in self._cm.env["nodes"]:
            should_be = self._cm.expected_status[node]
            rc = self._cm.test_node_cm(node)

            if rc > 0:
                if should_be == "down":
                    down_are_up += 1

                if rc == 1:
                    unstable_list.append(node)

            elif should_be == "up":
                up_are_down += 1

        if len(unstable_list) > 0:
            passed = False
            self._cm.log(f"Cluster is not stable: {len(unstable_list)} (of "
                         f"{self._cm.upcount()}): {unstable_list!r}")

        if up_are_down > 0:
            passed = False
            self._cm.log(f"{up_are_down} (of {len(self._cm.env['nodes'])}) nodes "
                         "expected to be up were down.")

        if down_are_up > 0:
            passed = False
            self._cm.log(f"{down_are_up} (of {len(self._cm.env['nodes'])}) nodes "
                         "expected to be down were up.")

        return passed

    def is_applicable(self):
        """Return True if this audit is applicable in the current test configuration."""
        # @TODO Due to long-ago refactoring, this name test would never match,
        # so this audit (and those derived from it) would never run.
        # Uncommenting the next lines fixes the name test, but that then
        # exposes pre-existing bugs that need to be fixed.
        # if self._cm["Name"] == "crm-corosync":
        #     return True
        return False


class CIBAudit(ClusterAudit):
    """Audit the CIB by verifying that it is identical across cluster nodes."""

    def __init__(self, cm):
        """
        Create a new CIBAudit instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        ClusterAudit.__init__(self, cm)
        self.name = "CibAudit"

    def __call__(self):
        passed = True
        ccm_partitions = self._cm.find_partitions()

        if not ccm_partitions:
            self.debug("\tNo partitions to audit")
            return passed

        for partition in ccm_partitions:
            self.debug(f"\tAuditing CIB consistency for: {partition}")

            if self._audit_cib_contents(partition) == 0:
                passed = False

        return passed

    def _audit_cib_contents(self, hostlist):
        """Perform the CIB audit on the given hosts."""
        passed = True
        node0 = None
        node0_xml = None

        partition_hosts = hostlist.split()
        for node in partition_hosts:
            node_xml = self._store_remote_cib(node, node0)

            if node_xml is None:
                self._cm.log(f"Could not perform audit: No configuration from {node}")
                passed = False

            elif node0 is None:
                node0 = node
                node0_xml = node_xml

            elif node0_xml is None:
                self._cm.log(f"Could not perform audit: No configuration from {node0}")
                passed = False

            else:
                (rc, result) = self._cm.rsh(
                    node0, f"crm_diff -VV -cf --new {node_xml} --original {node0_xml}", verbose=1)

                if rc != 0:
                    self._cm.log(f"Diff between {node0_xml} and {node_xml} failed: {rc}")
                    passed = False

                for line in result:
                    if not re.search("<diff/>", line):
                        passed = False
                        self.debug(f"CibDiff[{node0}-{node}]: {line}")
                    else:
                        self.debug(f"CibDiff[{node0}-{node}] Ignoring: {line}")

        return passed

    def _store_remote_cib(self, node, target):
        """
        Store a copy of the given node's CIB on the given target node.

        If no target is given, store the CIB on the given node.
        """
        filename = f"/tmp/ctsaudit.{node}.xml"

        if not target:
            target = node

        (rc, lines) = self._cm.rsh(node, self._cm["CibQuery"], verbose=1)
        if rc != 0:
            self._cm.log("Could not retrieve configuration")
            return None

        self._cm.rsh("localhost", f"rm -f {filename}")
        for line in lines:
            self._cm.rsh("localhost", f"echo \'{line[:-1]}\' >> {filename}", verbose=0)

        if self._cm.rsh.copy(filename, f"root@{target}:{filename}", silent=True) != 0:
            self._cm.log("Could not store configuration")
            return None

        return filename

    def is_applicable(self):
        """Return True if this audit is applicable in the current test configuration."""
        # @TODO Due to long-ago refactoring, this name test would never match,
        # so this audit (and those derived from it) would never run.
        # Uncommenting the next lines fixes the name test, but that then
        # exposes pre-existing bugs that need to be fixed.
        # if self._cm["Name"] == "crm-corosync":
        #     return True
        return False


class PartitionAudit(ClusterAudit):
    """
    Audit each partition in a cluster to verify a variety of conditions.

    Check that:

    * The number of partitions and the nodes in each is as expected
    * Each node is active when it should be active and inactive when it
      should be inactive
    * The status and epoch of each node is as expected
    * A partition has quorum
    * A partition has a DC when expected
    """

    def __init__(self, cm):
        """
        Create a new PartitionAudit instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        ClusterAudit.__init__(self, cm)
        self.name = "PartitionAudit"

        self._node_epoch = {}
        self._node_state = {}
        self._node_quorum = {}

    def __call__(self):
        passed = True
        ccm_partitions = self._cm.find_partitions()

        if not ccm_partitions:
            return passed

        self._cm.cluster_stable(double_check=True)

        if len(ccm_partitions) != self._cm.partitions_expected:
            self._cm.log(f"ERROR: {len(ccm_partitions)} cluster partitions detected:")
            passed = False

            for partition in ccm_partitions:
                self._cm.log(f"\t {partition}")

        for partition in ccm_partitions:
            if self._audit_partition(partition) == 0:
                passed = False

        return passed

    def _trim_string(self, avalue):
        """Remove the last character from a multi-character string."""
        if not avalue:
            return None

        if len(avalue) > 1:
            return avalue[:-1]

        return avalue

    def _trim2int(self, avalue):
        """Remove the last character from a multi-character string and convert the result to an int."""
        trimmed = self._trim_string(avalue)
        if trimmed:
            return int(trimmed)

        return None

    def _audit_partition(self, partition):
        """Perform the audit of a single partition."""
        passed = True
        dc_found = []
        dc_allowed_list = []
        lowest_epoch = None
        node_list = partition.split()

        self.debug(f"Auditing partition: {partition}")
        for node in node_list:
            if self._cm.expected_status[node] != "up":
                self._cm.log(f"Warn: Node {node} appeared out of nowhere")
                self._cm.expected_status[node] = "up"
                # not in itself a reason to fail the audit (not what we're
                #  checking for in this audit)

            (_, out) = self._cm.rsh(node, self._cm["StatusCmd"] % node, verbose=1)
            self._node_state[node] = out[0].strip()

            (_, out) = self._cm.rsh(node, self._cm["EpochCmd"], verbose=1)
            self._node_epoch[node] = out[0].strip()

            (_, out) = self._cm.rsh(node, self._cm["QuorumCmd"], verbose=1)
            self._node_quorum[node] = out[0].strip()

            self.debug(f"Node {node}: {self._node_state[node]} - {self._node_epoch[node]} - {self._node_quorum[node]}.")
            self._node_state[node] = self._trim_string(self._node_state[node])
            self._node_epoch[node] = self._trim2int(self._node_epoch[node])
            self._node_quorum[node] = self._trim_string(self._node_quorum[node])

            if not self._node_epoch[node]:
                self._cm.log(f"Warn: Node {node} disappeared: can't determine epoch")
                self._cm.expected_status[node] = "down"
                # not in itself a reason to fail the audit (not what we're
                #  checking for in this audit)
            elif lowest_epoch is None or self._node_epoch[node] < lowest_epoch:
                lowest_epoch = self._node_epoch[node]

        if not lowest_epoch:
            self._cm.log(f"Lowest epoch not determined in {partition}")
            passed = False

        for node in node_list:
            if self._cm.expected_status[node] != "up":
                continue

            if self._cm.is_node_dc(node, self._node_state[node]):
                dc_found.append(node)
                if self._node_epoch[node] == lowest_epoch:
                    self.debug(f"{node}: OK")
                elif not self._node_epoch[node]:
                    self.debug(f"Check on {node} ignored: no node epoch")
                elif not lowest_epoch:
                    self.debug(f"Check on {node} ignored: no lowest epoch")
                else:
                    self._cm.log(f"DC {node} is not the oldest node "
                                 f"({self._node_epoch[node]} vs. {lowest_epoch})")
                    passed = False

        if not dc_found:
            self._cm.log(f"DC not found on any of the {len(dc_allowed_list)} allowed "
                         f"nodes: {dc_allowed_list} (of {node_list})")

        elif len(dc_found) > 1:
            self._cm.log(f"{len(dc_found)} DCs ({dc_found}) found in cluster partition: {node_list}")
            passed = False

        if not passed:
            for node in node_list:
                if self._cm.expected_status[node] == "up":
                    self._cm.log(f"epoch {self._node_epoch[node]} : {self._node_state[node]}")

        return passed

    def is_applicable(self):
        """Return True if this audit is applicable in the current test configuration."""
        # @TODO Due to long-ago refactoring, this name test would never match,
        # so this audit (and those derived from it) would never run.
        # Uncommenting the next lines fixes the name test, but that then
        # exposes pre-existing bugs that need to be fixed.
        # if self._cm["Name"] == "crm-corosync":
        #     return True
        return False


# pylint: disable=invalid-name
def audit_list(cm):
    """Return a list of instances of applicable audits that can be performed."""
    result = []

    for auditclass in [DiskAudit, FileAudit, LogAudit, ControllerStateAudit,
                       PartitionAudit, PrimitiveAudit, GroupAudit, CloneAudit,
                       ColocationAudit, CIBAudit]:
        a = auditclass(cm)
        if a.is_applicable():
            result.append(a)

    return result



"""Toggle nodes in and out of maintenance mode."""

__all__ = ["MaintenanceMode"]
__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import re

from pacemaker._cts.audits import AuditResource
from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.starttest import StartTest
from pacemaker._cts.timer import Timer

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._rsh is callable.
# pylint: disable=not-callable


class MaintenanceMode(CTSTest):
    """Toggle nodes in and ount of maintenance mode."""

    def __init__(self, cm):
        """
        Create a new MaintenanceMode instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.benchmark = True
        self.name = "MaintenanceMode"

        self._action = "asyncmon"
        self._rid = "maintenanceDummy"
        self._start = StartTest(cm)
        self._startall = SimulStartLite(cm)

    def _toggle_maintenance_mode(self, node, enabled):
        """Toggle maintenance mode on the given node."""
        pats = [
            self.templates["Pat:DC_IDLE"]
        ]

        if enabled:
            action = "On"
        else:
            action = "Off"

        # fail the resource right after turning Maintenance mode on
        # verify it is not recovered until maintenance mode is turned off
        if enabled:
            pats.append(self.templates["Pat:RscOpFail"] % (self._action, self._rid))
        else:
            pats.extend([
                self.templates["Pat:RscOpOK"] % ("stop", self._rid),
                self.templates["Pat:RscOpOK"] % ("start", self._rid)
            ])

        watch = self.create_watch(pats, 60)
        watch.set_watch()

        self.debug(f"Turning maintenance mode {action}")
        self._rsh(node, self.templates[f"MaintenanceMode{action}"])

        if enabled:
            self._rsh(node, f"crm_resource -V -F -r {self._rid} -H {node} &>/dev/null")

        with Timer(self._logger, self.name, f"recover{action}"):
            watch.look_for_all()

        if watch.unmatched:
            self.debug(f"Failed to find patterns when turning maintenance mode {action}")
            return repr(watch.unmatched)

        return ""

    def _insert_maintenance_dummy(self, node):
        """Create a dummy resource on the given node."""
        pats = [
            f"{node}.*" + (self.templates["Pat:RscOpOK"] % ("start", self._rid))
        ]

        watch = self.create_watch(pats, 60)
        watch.set_watch()

        self._cm.add_dummy_rsc(node, self._rid)

        with Timer(self._logger, self.name, "addDummy"):
            watch.look_for_all()

        if watch.unmatched:
            self.debug("Failed to find patterns when adding maintenance dummy resource")
            return repr(watch.unmatched)

        return ""

    def _remove_maintenance_dummy(self, node):
        """Remove the previously created dummy resource on the given node."""
        pats = [
            self.templates["Pat:RscOpOK"] % ("stop", self._rid)
        ]

        watch = self.create_watch(pats, 60)
        watch.set_watch()
        self._cm.remove_dummy_rsc(node, self._rid)

        with Timer(self._logger, self.name, "removeDummy"):
            watch.look_for_all()

        if watch.unmatched:
            self.debug("Failed to find patterns when removing maintenance dummy resource")
            return repr(watch.unmatched)

        return ""

    def _managed_rscs(self, node):
        """Return a list of all resources managed by the cluster."""
        rscs = []
        (_, lines) = self._rsh(node, "crm_resource -c", verbose=1)

        for line in lines:
            if re.search("^Resource", line):
                tmp = AuditResource(self._cm, line)

                if tmp.managed:
                    rscs.append(tmp.id)

        return rscs

    def _verify_resources(self, node, rscs, managed):
        """Verify that all resources are managed or unmanaged as expected."""
        managed_rscs = rscs
        managed_str = "managed"

        if not managed:
            managed_str = "unmanaged"

        (_, lines) = self._rsh(node, "crm_resource -c", verbose=1)
        for line in lines:
            if re.search("^Resource", line):
                tmp = AuditResource(self._cm, line)

                if managed and not tmp.managed:
                    continue

                if not managed and tmp.managed:
                    continue

                if managed_rscs.count(tmp.id):
                    managed_rscs.remove(tmp.id)

        if not managed_rscs:
            self.debug(f"Found all {managed_str} resources on {node}")
            return True

        self._logger.log(f"Could not find all {managed_str} resources on {node}. {managed_rscs}")
        return False

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")
        verify_managed = False
        verify_unmanaged = False
        fail_pat = ""

        if not self._startall(None):
            return self.failure("Setup failed")

        # get a list of all the managed resources. We use this list
        # after enabling maintenance mode to verify all managed resources
        # become un-managed.  After maintenance mode is turned off, we use
        # this list to verify all the resources become managed again.
        managed_rscs = self._managed_rscs(node)
        if not managed_rscs:
            self._logger.log(f"No managed resources on {node}")
            return self.skipped()

        # insert a fake resource we can fail during maintenance mode
        # so we can verify recovery does not take place until after maintenance
        # mode is disabled.
        fail_pat += self._insert_maintenance_dummy(node)

        # toggle maintenance mode ON, then fail dummy resource.
        fail_pat += self._toggle_maintenance_mode(node, True)

        # verify all the resources are now unmanaged
        if self._verify_resources(node, managed_rscs, False):
            verify_unmanaged = True

        # Toggle maintenance mode  OFF, verify dummy is recovered.
        fail_pat += self._toggle_maintenance_mode(node, False)

        # verify all the resources are now managed again
        if self._verify_resources(node, managed_rscs, True):
            verify_managed = True

        # Remove our maintenance dummy resource.
        fail_pat += self._remove_maintenance_dummy(node)

        self._cm.cluster_stable()

        if fail_pat != "":
            return self.failure(f"Unmatched patterns: {fail_pat}")

        if not verify_unmanaged:
            return self.failure("Failed to verify resources became unmanaged during maintenance mode")

        if not verify_managed:
            return self.failure("Failed to verify resources switched back to managed after disabling maintenance mode")

        return self.success()

    @property
    def errors_to_ignore(self):
        """Return a list of errors which should be ignored."""
        return [
            f"Updating failcount for {self._rid}",
            fr"schedulerd.*: Recover\s+{self._rid}\s+\(.*\)",
            r"Unknown operation: fail",
            self.templates["Pat:RscOpOK"] % (self._action, self._rid),
            f"(ERROR|error).*: Action {self._rid}_{self._action}_0 .* initiated outside of a transition",
        ]



"""A module for managing and communicating with external processes."""

__all__ = ["killall", "exit_if_proc_running", "pipe_communicate", "stdout_from_command"]
__copyright__ = "Copyright 2009-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+)"

import subprocess
import sys

import psutil

from pacemaker.exitstatus import ExitStatus


def killall(process_names, terminate=False):
    """Kill all instances of every process in a list."""
    if not process_names:
        return

    if not isinstance(process_names, list):
        process_names = [process_names]

    procs = []
    for proc in psutil.process_iter(["name"]):
        if proc.info["name"] in process_names:
            procs.append(proc)

    if terminate:
        for proc in procs:
            proc.terminate()
        _, alive = psutil.wait_procs(procs, timeout=3)
        procs = alive

    for proc in procs:
        proc.kill()


def is_proc_running(process_name):
    """Check whether a process with a given name is running."""
    for proc in psutil.process_iter(["name"]):
        if proc.info["name"] == process_name:
            return True
    return False


def exit_if_proc_running(process_name):
    """Exit with error if a given process is running."""
    if is_proc_running(process_name):
        print(f"Error: {process_name} is already running!")
        print(f"Run {sys.argv[0]} only when the cluster is stopped.")
        sys.exit(ExitStatus.ERROR)


def pipe_communicate(pipes, check_stderr=False, stdin=None):
    """Get text output from pipes."""
    if stdin is not None:
        pipe_outputs = pipes.communicate(input=stdin.encode())
    else:
        pipe_outputs = pipes.communicate()

    output = pipe_outputs[0].decode(sys.stdout.encoding)
    if check_stderr:
        output += pipe_outputs[1].decode(sys.stderr.encoding)

    return output


def stdout_from_command(args):
    """Execute command and return its standard output."""
    with subprocess.Popen(args, stdout=subprocess.PIPE) as p:
        p.wait()
        return pipe_communicate(p).split("\n")



"""Start and stop a remote node."""

__all__ = ["RemoteBasic"]
__copyright__ = "Copyright 2000-2024 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.tests.remotedriver import RemoteDriver


class RemoteBasic(RemoteDriver):
    """Start and stop a remote node."""

    def __init__(self, cm):
        """
        Create a new RemoteBasic instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        RemoteDriver.__init__(self, cm)

        self.name = "RemoteBasic"

    def __call__(self, node):
        """Perform this test."""
        if not self.start_new_test(node):
            return self.failure(self.fail_string)

        self.test_attributes(node)
        self.cleanup_metal(node)

        self.debug("Waiting for the cluster to recover")
        self._cm.cluster_stable()
        if self.failed:
            return self.failure(self.fail_string)

        return self.success()
